PYE_VERSION   = " V2.72 "
try:
    import usys as sys
except:
    import sys
import gc
if sys.implementation.name == "micropython":
    is_micropython = True
    import uos as os
    from uio import StringIO
elif sys.implementation.name == "circuitpython":
    is_micropython = True
    import os
    from io import StringIO
else:
    is_micropython = False
    const = lambda x:x
    import os
    from _io import StringIO
from re import compile as re_compile
import time
KEY_NONE      = const(0x00)
KEY_UP        = const(0x0b)
KEY_DOWN      = const(0x0d)
KEY_LEFT      = const(0x1f)
KEY_RIGHT     = const(0x1e)
KEY_HOME      = const(0x10)
KEY_END       = const(0x03)
KEY_PGUP      = const(0xfff1)
KEY_PGDN      = const(0xfff2)
KEY_WORD_LEFT = const(0xfff3)
KEY_WORD_RIGHT= const(0xfff4)
KEY_SHIFT_UP  = const(0xfff5)
KEY_ALT_UP    = const(0xffea)
KEY_SHIFT_DOWN= const(0xfff6)
KEY_ALT_DOWN  = const(0xffeb)
KEY_SHIFT_LEFT= const(0xfff0)
KEY_ALT_LEFT  = const(0xffe9)
KEY_SHIFT_CTRL_LEFT= const(0xffed)
KEY_SHIFT_RIGHT= const(0xffef)
KEY_ALT_RIGHT = const(0xffe8)
KEY_SHIFT_CTRL_RIGHT= const(0xffec)
KEY_QUIT      = const(0x11)
KEY_FORCE_QUIT = const(0xffe6)
KEY_ENTER     = const(0x0a)
KEY_BACKSPACE = const(0x08)
KEY_DELETE    = const(0x7f)
KEY_DEL_WORD  = const(0xfff7)
KEY_DEL_LINE  = const(0xffe7)
KEY_WRITE     = const(0x13)
KEY_TAB       = const(0x09)
KEY_BACKTAB   = const(0x15)
KEY_FIND      = const(0x06)
KEY_GOTO      = const(0x07)
KEY_MOUSE     = const(0x1b)
KEY_SCRLUP    = const(0x1c)
KEY_SCRLDN    = const(0x1d)
KEY_FIND_AGAIN= const(0x0e)
KEY_REDRAW    = const(0x05)
KEY_UNDO      = const(0x1a)
KEY_REDO      = const(0xffee)
KEY_CUT       = const(0x18)
KEY_PASTE     = const(0x16)
KEY_COPY      = const(0x04)
KEY_FIRST     = const(0x14)
KEY_LAST      = const(0x02)
KEY_REPLC     = const(0x12)
KEY_TOGGLE    = const(0x01)
KEY_GET       = const(0x0f)
KEY_MARK      = const(0x0c)
KEY_NEXT      = const(0x17)
KEY_COMMENT   = const(0xfffc)
KEY_MATCH     = const(0xfffd)
KEY_INDENT    = const(0xfffe)
KEY_DEDENT    = const(0xffff)
class Editor:
    KEYMAP = {
    "\x1b[A" : KEY_UP,
    "\x1b[1;2A": KEY_SHIFT_UP,
    "\x1b[1;3A": KEY_ALT_UP,
    "\x1b[B" : KEY_DOWN,
    "\x1b[1;2B": KEY_SHIFT_DOWN,
    "\x1b[1;3B": KEY_ALT_DOWN,
    "\x1b[D" : KEY_LEFT,
    "\x1b[1;2D": KEY_SHIFT_LEFT,
    "\x1b[1;6D": KEY_SHIFT_CTRL_LEFT,
    "\x1b[1;3D": KEY_ALT_LEFT,
    "\x1b[C" : KEY_RIGHT,
    "\x1b[1;2C": KEY_SHIFT_RIGHT,
    "\x1b[1;6C": KEY_SHIFT_CTRL_RIGHT,
    "\x1b[1;3C": KEY_ALT_RIGHT,
    "\x1b[H" : KEY_HOME,
    "\x1bOH" : KEY_HOME,
    "\x1b[1~": KEY_HOME,
    "\x1b[F" : KEY_END,
    "\x1bOF" : KEY_END,
    "\x1b[4~": KEY_END,
    "\x1b[5~": KEY_PGUP,
    "\x1b[6~": KEY_PGDN,
    "\x1b[5;5~": KEY_PGUP,
    "\x1b[6;5~": KEY_PGDN,
    "\x1b[1;5D": KEY_WORD_LEFT,
    "\x1b[1;5C": KEY_WORD_RIGHT,
    "\x03"   : KEY_COPY,
    "\r"     : KEY_ENTER,
    "\x7f"   : KEY_BACKSPACE,
    "\x1b[3~": KEY_DELETE,
    "\x1b[Z" : KEY_BACKTAB,
    "\x19"   : KEY_REDO,
    "\x08"   : KEY_REPLC,
    "\x12"   : KEY_REPLC,
    "\x11"   : KEY_QUIT,
    "\x1b"   : KEY_QUIT,
    "\n"     : KEY_ENTER,
    "\x13"   : KEY_WRITE,
    "\x06"   : KEY_FIND,
    "\x0e"   : KEY_FIND_AGAIN,
    "\x07"   : KEY_GOTO,
    "\x05"   : KEY_REDRAW,
    "\x1a"   : KEY_UNDO,
    "\x09"   : KEY_TAB,
    "\x15"   : KEY_BACKTAB,
    "\x18"   : KEY_CUT,
    "\x16"   : KEY_PASTE,
    "\x04"   : KEY_COPY,
    "\x0c"   : KEY_MARK,
    "\x00"   : KEY_MARK,
    "\x14"   : KEY_FIRST,
    "\x02"   : KEY_LAST,
    "\x01"   : KEY_TOGGLE,
    "\x17"   : KEY_NEXT,
    "\x0f"   : KEY_GET,
    "\x10"   : KEY_COMMENT,
    "\x1f"   : KEY_COMMENT,
    "\x1b[1;5A": KEY_SCRLUP,
    "\x1b[1;5B": KEY_SCRLDN,
    "\x1b[1;5H": KEY_FIRST,
    "\x1b[1;5F": KEY_LAST,
    "\x1b[3;5~": KEY_DEL_WORD,
    "\x1b[3;2~": KEY_DEL_LINE,
    "\x0b"   : KEY_MATCH,
    "\x1b[M" : KEY_MOUSE,
    }
    TERMCMD = [
        "\x1b[{row};{col}H",
        "\x1b[0K",
        "\x1b[?25h",
        "\x1b[?25l",
        "\x1b[0m",
        "\x1b[1;37;46m",
        "\x1b[43m",
        '\x1b[?9h',
        '\x1b[?9l',
        "\x1bM",
        "\n",
        '\x1b[1;{stop}r',
        '\x1b[r',
        "\b",
        "{chd}{file} Row: {row}/{total} Col: {col}  {msg}",
        "{chd}{file} {row}:{col}  {msg}",
    ]
    yank_buffer = []
    find_pattern = ""
    case = "n"
    autoindent = "y"
    replc_pattern = ""
    comment_char = "\x23 "
    word_char = "_\\"
    file_char = "_.-"
    match_span = 50
    def __init__(self, tab_size, undo_limit, io_device):
        self.top_line = self.cur_line = self.row = self.vcol = self.col = self.margin = 0
        self.tab_size = tab_size
        self.changed = ''
        self.hash = 0
        self.message = self.fname = ""
        self.content = [""]
        self.undo = []
        self.undo_limit = undo_limit
        self.redo = []
        self.clear_mark()
        self.write_tabs = "n"
        self.work_dir = os.getcwd()
        self.io_device = io_device
        self.wr = io_device.wr
        self.is_dir = False
        self.key_max = 0
        for _ in Editor.KEYMAP.keys():
            self.key_max = max(self.key_max, len(_))
    def goto(self, row, col):
        self.wr(Editor.TERMCMD[0].format(row=row + 1, col=col + 1))
    def clear_to_eol(self):
        self.wr(Editor.TERMCMD[1])
    def cursor(self, onoff):
        self.wr(Editor.TERMCMD[2] if onoff else Editor.TERMCMD[3])
    def hilite(self, mode):
        if mode == 1:
            self.wr(Editor.TERMCMD[5])
        elif mode == 2:
            self.wr(Editor.TERMCMD[6])
        else:
            self.wr(Editor.TERMCMD[4])
    def mouse_reporting(self, onoff):
        self.wr(Editor.TERMCMD[7] if onoff else Editor.TERMCMD[8])
    def scroll_region(self, stop):
        self.wr(Editor.TERMCMD[11].format(stop=stop) if stop else Editor.TERMCMD[12])
    def scroll_up(self, scrolling):
        if Editor.TERMCMD[9]:
            Editor.scrbuf[scrolling:] = Editor.scrbuf[:-scrolling]
            Editor.scrbuf[:scrolling] = [''] * scrolling
            self.goto(0, 0)
            self.wr(Editor.TERMCMD[9] * scrolling)
    def scroll_down(self, scrolling):
        if Editor.TERMCMD[10]:
            Editor.scrbuf[:-scrolling] = Editor.scrbuf[scrolling:]
            Editor.scrbuf[-scrolling:] = [''] * scrolling
            self.goto(Editor.height - 1, 0)
            self.wr(Editor.TERMCMD[10] * scrolling)
    def redraw(self, flag):
        self.cursor(False)
        Editor.height, Editor.width = self.io_device.get_screen_size()
        Editor.height -= 1
        Editor.scrbuf = [(False,"\x00")] * Editor.height
        self.row = min(Editor.height - 1, self.row)
        self.scroll_region(Editor.height)
        self.mouse_reporting(True)
        if flag:
            self.message = PYE_VERSION
        if is_micropython:
            gc.collect()
            if flag:
                self.message += "{} Bytes Memory available".format(gc.mem_free())
        self.changed = '' if self.hash == self.hash_buffer() else '*'
    def get_input(self):
        while True:
            in_buffer = self.io_device.rd()
            if in_buffer[0] == '\x1b':
                while True:
                    in_buffer += self.io_device.rd()
                    c = in_buffer[-1]
                    if c == '~' or (c.isalpha() and len(in_buffer) > 2):
                        break
                    if len(in_buffer) == 2 and c.isalpha() and c != 'O':
                        in_buffer = chr(ord(in_buffer[1]) & 0x1f)
                        break
                    if len(in_buffer) >= self.key_max:
                        break
                    if in_buffer == '\x1b\x1b':
                        in_buffer = '\x1b'
                        break
            if in_buffer in Editor.KEYMAP:
                c = Editor.KEYMAP[in_buffer]
                if c != KEY_MOUSE:
                    return c, None
                else:
                    mouse_fct = ord(self.io_device.rd_raw())
                    mouse_x = ord(self.io_device.rd_raw()) - 33
                    mouse_y = ord(self.io_device.rd_raw()) - 33
                    if mouse_fct == 0x61:
                        return KEY_SCRLDN, 3
                    elif mouse_fct == 0x60:
                        return KEY_SCRLUP, 3
                    else:
                        return KEY_MOUSE, [mouse_x, mouse_y, mouse_fct]
            elif ord(in_buffer[0]) >= 32:
                return KEY_NONE, in_buffer
    def display_window(self):
        self.cur_line = min(self.total_lines - 1, max(self.cur_line, 0))
        self.vcol = max(0, min(self.col, len(self.content[self.cur_line])))
        if self.vcol >= Editor.width + self.margin:
            self.margin = self.vcol - Editor.width + (Editor.width >> 2)
        elif self.vcol < self.margin:
            self.margin = max(self.vcol - (Editor.width >> 2), 0)
        if not (self.top_line <= self.cur_line < self.top_line + Editor.height):
            self.top_line = max(self.cur_line - self.row, 0)
        self.row = self.cur_line - self.top_line
        self.cursor(False)
        line = self.top_line
        if self.mark is None:
            flag = 0
        else:
            start_line, start_col, end_line, end_col = self.mark_range()
            start_col = max(start_col - self.margin, 0)
            end_col = max(end_col - self.margin, 0)
        for c in range(Editor.height):
            if line == self.total_lines:
                if Editor.scrbuf[c] != (False,''):
                    self.goto(c, 0)
                    self.clear_to_eol()
                    Editor.scrbuf[c] = (False,'')
            else:
                if self.mark is not None:
                    flag = ((start_line <= line < end_line) +
                            ((start_line == line) << 1) +
                            (((end_line - 1) == line) << 2))
                l = (flag,
                     self.content[line][self.margin:self.margin + Editor.width])
                if (flag and line == self.cur_line) or l != Editor.scrbuf[c]:
                    self.goto(c, 0)
                    if flag == 0:
                        self.wr(l[1])
                    elif flag == 7:
                        self.wr(l[1][:start_col])
                        self.hilite(2)
                        self.wr(l[1][start_col:end_col])
                        self.hilite(0)
                        self.wr(l[1][end_col:])
                    elif flag == 3:
                        self.wr(l[1][:start_col])
                        self.hilite(2)
                        self.wr(l[1][start_col:])
                        self.wr(' ')
                        self.hilite(0)
                    elif flag == 5:
                        self.hilite(2)
                        self.wr(l[1][:end_col])
                        self.hilite(0)
                        self.wr(l[1][end_col:])
                    else:
                        self.hilite(2)
                        self.wr(l[1])
                        self.wr(' ')
                        self.hilite(0)
                    if len(l[1]) < Editor.width:
                        self.clear_to_eol()
                    Editor.scrbuf[c] = l
                line += 1
        self.goto(Editor.height, 0)
        self.hilite(1)
        self.wr(Editor.TERMCMD[14 if Editor.width > 40 else 15].format(
            chd=self.changed, file=self.fname, row=self.cur_line + 1,
            total=self.total_lines, col=self.vcol + 1, msg=self.message)[:self.width - 1])
        self.clear_to_eol()
        self.hilite(0)
        self.goto(self.row, self.vcol - self.margin)
        self.cursor(True)
    def spaces(self, line, pos = None):
        return (len(line) - len(line.lstrip(" ")) if pos is None else
                len(line[:pos]) - len(line[:pos].rstrip(" ")))
    def mark_range(self):
        if self.mark_order(self.cur_line, self.col) >= 0:
            return (self.mark[0], self.mark[1], self.cur_line + 1, self.col)
        else:
            return (self.cur_line, self.col, self.mark[0] + 1, self.mark[1])
    def mark_order(self, line, col):
        return col - self.mark[1] if self.mark[0] == line else line - self.mark[0]
    def line_range(self):
        res = self.mark_range()
        return (res[0], res[2]) if res[3] > 0 else (res[0], res[2] - 1)
    def line_edit(self, prompt, default, zap=None):
        push_msg = lambda msg: self.wr(msg + Editor.TERMCMD[13] * len(msg))
        self.goto(Editor.height, 0)
        self.hilite(1)
        self.wr(prompt)
        self.wr(default)
        self.clear_to_eol()
        res = default
        pos = len(res)
        del_all = True
        mouse_last = None
        while True:
            key, char = self.get_input()
            if key == KEY_NONE:
                if len(prompt) + len(res) < self.width - 2:
                    res = res[:pos] + char + res[pos:]
                    self.wr(res[pos])
                    pos += len(char)
                    push_msg(res[pos:])
            elif key in (KEY_ENTER, KEY_TAB):
                self.hilite(0)
                return res
            elif key in (KEY_QUIT, KEY_COPY):
                self.hilite(0)
                return None
            elif key == KEY_LEFT:
                if pos > 0:
                    self.wr(Editor.TERMCMD[13])
                    pos -= 1
            elif key == KEY_RIGHT:
                if pos < len(res):
                    self.wr(res[pos])
                    pos += 1
            elif key == KEY_HOME:
                self.wr(Editor.TERMCMD[13] * pos)
                pos = 0
            elif key == KEY_END:
                self.wr(res[pos:])
                pos = len(res)
            elif key == KEY_DELETE:
                if del_all:
                    self.wr(Editor.TERMCMD[13] * pos)
                    self.wr(" " * len(res))
                    self.wr(Editor.TERMCMD[13] * len(res))
                    pos = 0
                    res = ""
                else:
                    if pos < len(res):
                        res = res[:pos] + res[pos+1:]
                        push_msg(res[pos:] + ' ')
            elif key == KEY_BACKSPACE:
                if pos > 0:
                    res = res[:pos-1] + res[pos:]
                    self.wr(Editor.TERMCMD[13])
                    pos -= 1
                    push_msg(res[pos:] + ' ')
            elif key == KEY_PASTE:
                res += self.getsymbol(self.content[self.cur_line], self.col, zap)[:Editor.width - pos - len(prompt) - 1]
                push_msg(res[pos:])
            elif key == KEY_MOUSE:
                if char[1] < Editor.height and (char[1] + self.top_line) < self.total_lines:
                    self.col = char[0] + self.margin
                    self.cur_line = char[1] + self.top_line
                    if (self.col, self.cur_line) != mouse_last:
                        mouse_last = (self.col, self.cur_line)
                        self.wr(Editor.TERMCMD[13] * pos)
                        self.wr(" " * len(res))
                        self.wr(Editor.TERMCMD[13] * len(res))
                        pos = 0
                        res = self.getsymbol(self.content[self.cur_line], self.col, zap)
                        push_msg(res)
                    else:
                        self.hilite(0)
                        return res
            del_all = False
    def getsymbol(self, s, pos, zap):
        if pos < len(s) and zap is not None:
            start = self.skip_while(s, pos, zap, -1)
            stop = self.skip_while(s, pos, zap, 1)
            return s[start+1:stop]
        else:
            return ''
    def issymbol(self, c, zap):
        return c.isalpha() or c.isdigit() or c in zap
    def skip_until(self, s, pos, zap, way):
        stop = -1 if way < 0 else len(s)
        while pos != stop and not self.issymbol(s[pos], zap):
            pos += way
        return pos
    def skip_while(self, s, pos, zap, way):
        stop = -1 if way < 0 else len(s)
        while pos != stop and self.issymbol(s[pos], zap):
            pos += way
        return pos
    def move_up(self):
        if self.cur_line > 0:
            self.cur_line -= 1
            if self.cur_line < self.top_line:
                self.scroll_up(1)
    def skip_up(self):
        if self.col == 0 and self.cur_line > 0:
            self.col = len(self.content[self.cur_line - 1])
            self.move_up()
            return True
        else:
            return False
    def move_left(self):
        self.col = self.vcol
        if not self.skip_up():
            self.col -= 1
    def move_down(self):
        if self.cur_line < self.total_lines - 1:
            self.cur_line += 1
            if self.cur_line == self.top_line + Editor.height:
                self.scroll_down(1)
    def skip_down(self, l):
        if self.col >= len(l) and self.cur_line < self.total_lines - 1:
            self.col = 0
            self.move_down()
            return True
        else:
            return False
    def move_right(self, l):
        if not self.skip_down(l):
            self.col += 1
    def find_in_file(self, pattern, col, end):
        Editor.find_pattern = pattern
        if Editor.case != "y":
            pattern = pattern.lower()
        try:
            rex = re_compile(pattern)
        except:
            self.message = "Invalid pattern: " + pattern
            return None
        start = self.cur_line
        if (col > len(self.content[start]) or
            (pattern[0] == '^' and col != 0)):
            start, col = start + 1, 0
        for line in range(start, end):
            l = self.content[line][col:]
            if Editor.case != "y":
                l = l.lower()
            match = rex.search(l)
            if match:
                self.cur_line = line
                if pattern[-1:] == "$" and match.group(0)[-1:] != "$":
                    self.col = col + len(l) - len(match.group(0))
                else:
                    self.col = col + l.find(match.group(0))
                return len(match.group(0))
            col = 0
        else:
            self.message = Editor.find_pattern + " not found (again)"
            return None
    def undo_add(self, lnum, text, key, span = 1, chain=False):
        if (len(self.undo) == 0 or key == KEY_NONE or
            self.undo[-1][3] != key or self.undo[-1][0] != lnum):
            self.changed = '*'
            if len(self.undo) >= self.undo_limit:
                del self.undo[0]
            self.undo.append([lnum, span, text, key, self.col, chain])
            self.redo = []
    def undo_redo(self, undo, redo):
        chain = True
        redo_start = len(redo)
        while len(undo) > 0 and chain:
            action = undo.pop()
            if not action[3] in (KEY_INDENT, KEY_DEDENT, KEY_COMMENT):
                self.cur_line = action[0]
            self.col = action[4]
            if len(redo) >= self.undo_limit:
                del redo[0]
            if action[1] >= 0:
                redo.append(action[0:1] + [len(action[2])] +
                    [self.content[action[0]:action[0] + action[1]]] + action[3:])
                if action[0] < self.total_lines:
                    self.content[action[0]:action[0] + action[1]] = action[2]
                else:
                    self.content += action[2]
            else:
                redo.append(action[0:1] + [1] +
                    [self.content[action[0]:action[0] - action[1] + 1]] + action[3:])
                del self.content[action[0]:action[0] - action[1]]
                self.content[action[0]] = action[2][0]
            chain = action[5]
        if (len(redo) - redo_start) > 0:
            redo[-1][5] = True
            redo[redo_start][5] = False
            self.total_lines = len(self.content)
            self.changed = '' if self.hash == self.hash_buffer() else '*'
            self.clear_mark()
    def set_mark(self, flag=999999999):
        if self.mark is None:
            self.mark = (self.cur_line, self.col)
        if self.mark_flag < flag:
            self.mark_flag = flag
    def check_mark(self):
        if self.mark is not None:
            self.mark_flag -= 1
            if self.mark_flag <= 0:
                self.clear_mark()
    def clear_mark(self):
        self.mark = None
        self.mark_flag = 0
        self.mouse_last = (0, 0, 0)
    def yank_mark(self):
        start_row, start_col, end_row, end_col = self.mark_range()
        Editor.yank_buffer = self.content[start_row:end_row]
        Editor.yank_buffer[-1] = Editor.yank_buffer[-1][:end_col]
        Editor.yank_buffer[0] = Editor.yank_buffer[0][start_col:]
    def delete_mark(self, yank):
        if yank:
            self.yank_mark()
        start_row, start_col, end_row, end_col = self.mark_range()
        self.undo_add(start_row, self.content[start_row:end_row], KEY_NONE, 1, False)
        self.content[start_row] = self.content[start_row][:start_col] + self.content[end_row - 1][end_col:]
        if start_row + 1 < end_row:
            del self.content[start_row + 1:end_row]
        self.col = start_col
        if self.content == []:
            self.content = [""]
            self.undo[-1][1] = 1
        self.total_lines = len(self.content)
        self.cur_line = start_row
        self.clear_mark()
    def handle_edit_keys(self, key, char):
        l = self.content[self.cur_line]
        if key == KEY_NONE:
            self.col = self.vcol
            if self.mark is not None:
                self.delete_mark(False)
                l = self.content[self.cur_line]
                chain = True
            else:
                chain = False
            self.undo_add(self.cur_line, [l], 0x20 if char == " " else 0x41, 1, chain)
            self.content[self.cur_line] = l[:self.col] + char + l[self.col:]
            self.col += len(char)
            return key
        elif key == KEY_SHIFT_CTRL_LEFT:
            self.set_mark()
            key = KEY_WORD_LEFT
        elif key == KEY_SHIFT_CTRL_RIGHT:
            self.set_mark()
            key = KEY_WORD_RIGHT
        elif key == KEY_MOUSE:
            if char[2] == 0x22:
                key = KEY_GET if self.is_dir else KEY_FIND
            elif char[1] < Editor.height:
                col = char[0] + self.margin
                line = char[1] + self.top_line
                if (col, line) == self.mouse_last[:2] and (time.time() - self.mouse_last[2]) < 2:
                    self.mouse_last = (0, 0, 0)
                    if self.mark is None and col < len(l) and self.issymbol(l[col], Editor.word_char): 
                            self.col = self.skip_while(l, col, Editor.word_char, -1) + 1
                            self.set_mark()
                            self.col = self.skip_while(l, self.col, Editor.word_char, 1)
                    else:
                        key = KEY_MARK
                else:
                    if self.mark is not None:
                        if self.mark_order(self.cur_line, self.col) * self.mark_order(line, col) < 0:
                            self.mark = self.cur_line, self.col
                    self.cur_line, self.col = line, col
                    self.mouse_last = (col, line, time.time())
        if key == KEY_DOWN:
             self.move_down()
        elif key == KEY_UP:
            self.move_up()
        elif key == KEY_LEFT:
            self.move_left()
        elif key == KEY_RIGHT:
            self.move_right(l)
        elif key == KEY_WORD_LEFT:
            self.col = self.vcol
            if self.skip_up():
                l = self.content[self.cur_line]
            pos = self.skip_until(l, self.col - 1, Editor.word_char, -1)
            self.col = self.skip_while(l, pos, Editor.word_char, -1) + 1
        elif key == KEY_WORD_RIGHT:
            if self.skip_down(l):
                l = self.content[self.cur_line]
            pos = self.skip_until(l, self.col, Editor.word_char, 1)
            self.col = self.skip_while(l, pos, Editor.word_char, 1)
        elif key == KEY_DELETE:
            self.col = self.vcol
            if self.mark is not None:
                self.delete_mark(False)
            elif self.col < len(l):
                self.undo_add(self.cur_line, [l], KEY_DELETE)
                self.content[self.cur_line] = l[:self.col] + l[self.col + 1:]
            elif (self.cur_line + 1) < self.total_lines:
                self.undo_add(self.cur_line, [l, self.content[self.cur_line + 1]], KEY_NONE)
                self.content[self.cur_line] = l + (
                    self.content.pop(self.cur_line + 1).lstrip()
                    if Editor.autoindent == "y" and self.col > 0
                    else self.content.pop(self.cur_line + 1))
                self.total_lines -= 1
        elif key == KEY_BACKSPACE:
            self.col = self.vcol
            if self.mark is not None:
                self.delete_mark(False)
            elif self.col > 0:
                self.undo_add(self.cur_line, [l], KEY_BACKSPACE)
                self.content[self.cur_line] = l[:self.col - 1] + l[self.col:]
                self.col -= 1
            elif self.cur_line > 0:
                self.undo_add(self.cur_line - 1, [self.content[self.cur_line - 1], l], KEY_NONE)
                self.col = len(self.content[self.cur_line - 1])
                self.content[self.cur_line - 1] += self.content.pop(self.cur_line)
                self.cur_line -= 1
                self.total_lines -= 1
        elif key == KEY_DEL_WORD:
            if self.col < len(l):
                pos = self.skip_while(l, self.col, Editor.word_char, 1)
                pos += self.spaces(l[pos:])
                if self.col < pos:
                    self.undo_add(self.cur_line, [l], KEY_DEL_WORD)
                    self.content[self.cur_line] = l[:self.col] + l[pos:]
        elif key == KEY_DEL_LINE:
            if self.cur_line < (self.total_lines - 1):
                self.undo_add(self.cur_line, [l, self.content[self.cur_line + 1]], KEY_NONE, 1)
            else:
                self.undo_add(self.cur_line, [l], KEY_NONE, 1)
            self.content.pop(self.cur_line)
            if self.content == []:
                self.content = [""]
        elif key == KEY_HOME:
            self.col = self.spaces(l) if self.col == 0 else 0
        elif key == KEY_END:
            ni = len(l.split(Editor.comment_char.strip())[0].rstrip())
            ns = self.spaces(l)
            self.col = ni if self.col >= len(l) and ni > ns else len(l)
        elif key == KEY_PGUP:
            self.cur_line -= Editor.height
        elif key == KEY_PGDN:
            self.cur_line += Editor.height
        elif key == KEY_FIND:
            pat = self.line_edit("Find: ", Editor.find_pattern, "_")
            if pat:
                self.clear_mark()
                self.find_in_file(pat, self.col + 1, self.total_lines)
                self.row = Editor.height >> 1
        elif key == KEY_FIND_AGAIN:
            if Editor.find_pattern:
                self.find_in_file(Editor.find_pattern, self.col + 1, self.total_lines)
                self.row = Editor.height >> 1
        elif key == KEY_GOTO:
            line = self.line_edit("Goto Line: ", "")
            if line:
                self.cur_line = int(line) - 1
                self.row = Editor.height >> 1
        elif key == KEY_FIRST:
            self.check_mark()
            self.cur_line = 0
        elif key == KEY_LAST:
            self.check_mark()
            self.cur_line = self.total_lines - 1
            self.row = Editor.height - 1
        elif key == KEY_TOGGLE:
            pat = self.line_edit("Autoindent {}, Search Case {}"
            ", Tabsize {}, Comment {}, Tabwrite {}: ".format(
            Editor.autoindent, Editor.case, self.tab_size, Editor.comment_char, self.write_tabs), "")
            try:
                res =  [i.lstrip().lower() for i in pat.split(",")]
                if res[0]: Editor.autoindent = 'y' if res[0][0] == 'y' else 'n'
                if res[1]: Editor.case     = 'y' if res[1][0] == 'y' else 'n'
                if res[2]: self.tab_size = int(res[2])
                if res[3]: Editor.comment_char = res[3]
                if res[4]: self.write_tabs = 'y' if res[4][0] == 'y' else 'n'
            except:
                pass
        elif key == KEY_SCRLUP:
            ni = 1 if char is None else 3
            if self.top_line > 0:
                self.top_line = max(self.top_line - ni, 0)
                self.cur_line = min(self.cur_line, self.top_line + Editor.height - 1)
                self.scroll_up(ni)
        elif key == KEY_SCRLDN:
            ni = 1 if char is None else 3
            if self.top_line + Editor.height < self.total_lines:
                self.top_line = min(self.top_line + ni, self.total_lines - 1)
                self.cur_line = max(self.cur_line, self.top_line)
                self.scroll_down(ni)
        elif key == KEY_MATCH:
            if self.col < len(l):
                brackets = "<{[()]}>"
                srch = l[self.col]
                i = brackets.find(srch)
                if i >= 0:
                    match = brackets[7 - i]
                    level = 0
                    way = 1 if i < 4 else -1
                    i = self.cur_line
                    c = self.col + way
                    lstop = (min(self.total_lines, i + Editor.match_span)
                             if way > 0 else
                             max(-1, i - Editor.match_span))
                    while i != lstop:
                        l = self.content[i]
                        cstop = len(l) if way > 0 else -1
                        if srch in l or match in l:
                            while c != cstop:
                                if l[c] == match:
                                    if level == 0:
                                        self.cur_line, self.col  = i, c
                                        return key
                                    else:
                                        level -= 1
                                elif l[c] == srch:
                                    level += 1
                                c += way
                        i += way
                        c = 0 if way > 0 else len(self.content[i]) - 1
                    self.message = "No match in {} lines".format(abs(lstop - self.cur_line))
        elif key == KEY_MARK:
            if self.mark is None: 
                self.set_mark()
                self.move_right(l)
            else:
                self.clear_mark()
        elif key == KEY_SHIFT_DOWN:
            self.set_mark()
            self.move_down()
        elif key == KEY_SHIFT_UP:
            self.set_mark()
            self.move_up()
        elif key == KEY_SHIFT_LEFT:
            self.set_mark()
            self.move_left()
        elif key == KEY_SHIFT_RIGHT:
            self.set_mark()
            self.move_right(l)
        elif key == KEY_ALT_LEFT:
            if self.col > 0:
                self.undo_add(self.cur_line, [l], KEY_ALT_LEFT)
                i = self.col
                self.content[self.cur_line] = l[:i - 1] + l[i] + l[i - 1] + l[i + 1:]
                self.move_left()
        elif key == KEY_ALT_RIGHT:
            if self.col < len(l):
                self.undo_add(self.cur_line, [l], KEY_ALT_RIGHT)
                i = self.col
                self.content[self.cur_line] = l[:i] + l[i + 1] + l[i] + l[i + 2:]
                self.move_right(l)
        elif key == KEY_ALT_UP:
            if self.mark is None:
                start_line = self.cur_line
                end_line = start_line + 1
            else:
                start_line, end_line = self.line_range()
                if start_line > 0:
                    self.mark = (self.mark[0] - 1, self.mark[1])
            if start_line > 0:
                self.undo_add(start_line - 1, self.content[start_line - 1:end_line],
                              KEY_NONE, end_line - start_line + 1)
                self.content[start_line - 1:end_line - 1], self.content[end_line - 1] = (
                    self.content[start_line:end_line], self.content[start_line - 1])
                self.move_up()
        elif key == KEY_ALT_DOWN:
            if self.mark is None:
                start_line = self.cur_line
                end_line = start_line + 1
            else:
                start_line, end_line = self.line_range()
                if end_line < self.total_lines:
                    self.mark = (self.mark[0] + 1, self.mark[1])
                    if self.cur_line == end_line == (self.total_lines - 1):
                        self.move_left()
            if end_line < self.total_lines:
                self.undo_add(start_line, self.content[start_line:end_line + 1],
                              KEY_NONE, end_line - start_line + 1)
                self.content[start_line + 1:end_line + 1], self.content[start_line] = (
                    self.content[start_line:end_line], self.content[end_line])
                self.move_down()
        elif key == KEY_ENTER:
            self.col = self.vcol
            self.clear_mark()
            self.undo_add(self.cur_line, [l], KEY_NONE, 2)
            self.content[self.cur_line] = l[:self.col]
            ni = 0
            if Editor.autoindent == "y":
                ni = min(self.spaces(l), self.col)
            self.cur_line += 1
            self.content[self.cur_line:self.cur_line] = [' ' * ni + l[self.col:]]
            self.total_lines += 1
            self.col = ni
        elif key == KEY_TAB:
            if self.mark is None:
                self.col = self.vcol
                self.undo_add(self.cur_line, [l], KEY_TAB)
                ni = self.tab_size - self.col % self.tab_size
                self.content[self.cur_line] = l[:self.col] + ' ' * ni + l[self.col:]
                self.col += ni
            else:
                lrange = self.line_range()
                self.undo_add(lrange[0], self.content[lrange[0]:lrange[1]], KEY_INDENT, lrange[1] - lrange[0])
                for i in range(lrange[0],lrange[1]):
                    if len(self.content[i]) > 0:
                        self.content[i] = ' ' * (self.tab_size - self.spaces(self.content[i]) % self.tab_size) + self.content[i]
        elif key == KEY_BACKTAB:
            if self.mark is None:
                self.col = self.vcol
                ni = min((self.col - 1) % self.tab_size + 1, self.spaces(l, self.col))
                if ni > 0:
                    self.undo_add(self.cur_line, [l], KEY_BACKTAB)
                    self.content[self.cur_line] = l[:self.col - ni] + l[self.col:]
                    self.col -= ni
            else:
                lrange = self.line_range()
                self.undo_add(lrange[0], self.content[lrange[0]:lrange[1]], KEY_DEDENT, lrange[1] - lrange[0])
                for i in range(lrange[0],lrange[1]):
                    ns = self.spaces(self.content[i])
                    if ns > 0:
                        self.content[i] = self.content[i][(ns - 1) % self.tab_size + 1:]
        elif key == KEY_REPLC:
            count = 0
            pat = self.line_edit("Replace: ", Editor.find_pattern, "_")
            if pat:
                rpat = self.line_edit("With: ", Editor.replc_pattern if Editor.replc_pattern else pat)
                if rpat is not None:
                    Editor.replc_pattern = rpat
                    q = ''
                    cur_line, cur_col = self.cur_line, self.col
                    if self.mark is not None:
                        (self.cur_line, self.col, end_line, end_col) = self.mark_range()
                    else:
                        end_line = self.total_lines
                        end_col = 999999
                    self.message = "Replace (yes/No/all/quit) ? "
                    chain = False
                    while True:
                        ni = self.find_in_file(pat, self.col, end_line)
                        if ni is not None and (self.cur_line != (end_line - 1) or self.col < end_col):
                            if q != 'a':
                                self.display_window()
                                key, char = self.get_input()
                                q = char.lower()
                            if q == 'q' or key == KEY_QUIT:
                                break
                            elif q in ('a','y'):
                                self.undo_add(self.cur_line, [self.content[self.cur_line]], KEY_NONE, 1, chain)
                                self.content[self.cur_line] = self.content[self.cur_line][:self.col] + rpat + self.content[self.cur_line][self.col + ni:]
                                self.col += len(rpat) + (ni == 0)
                                count += 1
                                chain = True
                            else:
                                 self.col += 1
                        else:
                            break
                    self.cur_line, self.col = cur_line, cur_col
                    self.message = "'{}' replaced {} times".format(pat, count)
        elif key == KEY_CUT:
            if self.mark is not None:
                self.delete_mark(True)
        elif key == KEY_COPY:
            if self.mark is not None:
                self.yank_mark()
                self.clear_mark()
        elif key == KEY_PASTE:
            if Editor.yank_buffer:
                self.col = self.vcol
                if self.mark is not None:
                    self.delete_mark(False)
                    chain = True
                else:
                    chain = False
                head, tail = Editor.yank_buffer[0], Editor.yank_buffer[-1]
                Editor.yank_buffer[0] = self.content[self.cur_line][:self.col] + Editor.yank_buffer[0]
                Editor.yank_buffer[-1] += self.content[self.cur_line][self.col:]
                ni = 1 if len(Editor.yank_buffer) <= 1 else 1 - len(Editor.yank_buffer)
                self.undo_add(self.cur_line, [self.content[self.cur_line]], KEY_NONE, ni, chain)
                self.content[self.cur_line:self.cur_line + 1] = Editor.yank_buffer
                Editor.yank_buffer[-1], Editor.yank_buffer[0] = tail, head
                self.total_lines = len(self.content)
        elif key == KEY_WRITE:
            fname = self.line_edit("Save File: ", self.fname if self.is_dir is False else "", Editor.file_char)
            if fname:
                if fname != self.fname:
                    try:
                        open(fname).close()
                        res = self.line_edit("The file exists! Overwrite (y/N)? ", "N")
                        if not res or res[0].upper() != 'Y':
                            return
                    except:
                        pass
                self.put_file(fname)
                self.fname = fname
                self.hash = self.hash_buffer()
                self.changed = ''
                self.is_dir = False
        elif key == KEY_UNDO:
            self.undo_redo(self.undo, self.redo)
        elif key == KEY_REDO:
            self.undo_redo(self.redo, self.undo)
        elif key == KEY_COMMENT:
            if self.mark is None:
                lrange = (self.cur_line, self.cur_line + 1)
            else:
                lrange = self.line_range()
            self.undo_add(lrange[0], self.content[lrange[0]:lrange[1]], KEY_COMMENT, lrange[1] - lrange[0])
            ni = len(Editor.comment_char)
            for i in range(lrange[0],lrange[1]):
                if self.content[i].strip() != "":
                    ns = self.spaces(self.content[i])
                    if self.content[i][ns:ns + ni] == Editor.comment_char:
                        self.content[i] = ns * " " + self.content[i][ns + ni:]
                    else:
                        self.content[i] = ns * " " + Editor.comment_char + self.content[i][ns:]
        elif key == KEY_REDRAW:
            self.redraw(True)
        return key
    def edit_loop(self):
        if not self.content:
            self.content = [""]
        self.total_lines = len(self.content)
        os.chdir(self.work_dir)
        self.redraw(self.message == "")
        while True:
            self.display_window()
            key, char = self.get_input()
            self.message = ''
            key = self.handle_edit_keys(key, char)
            if key == KEY_QUIT:
                if self.hash != self.hash_buffer():
                    res = self.line_edit("File changed! Quit (y/N/f)? ", "N")
                    if not res or res[0].upper() == 'N':
                        continue
                    if res[0].upper() == 'F':
                        key = KEY_FORCE_QUIT
                self.scroll_region(0)
                self.mouse_reporting(False)
                self.goto(Editor.height, 0)
                self.clear_to_eol()
                self.undo = []
                return key
            elif key == KEY_NEXT:
                return key
            elif key == KEY_GET:
                if self.mark is not None:
                    self.clear_mark()
                    self.display_window()
                return key
    def packtabs(self, s):
        sb = StringIO()
        for i in range(0, len(s), 8):
            c = s[i:i + 8]
            cr = c.rstrip(" ")
            if (len(c) - len(cr)) > 1:
                sb.write(cr + "\t")
            else:
                sb.write(c)
        return sb.getvalue()
    def hash_buffer(self):
        res = 0
        for line in self.content:
            res = ((res * 227 + 1) ^ hash(line)) & 0x3fffffff
        return res
    def get_file(self, fname):
        if fname:
            try:
                self.fname = fname
                if fname in ('.', '..') or (os.stat(fname)[0] & 0x4000):
                    os.chdir(fname)
                    self.work_dir = os.getcwd()
                    self.fname = "/" if self.work_dir == "/" else self.work_dir.split("/")[-1]
                    self.content = ["Directory '{}'".format(self.work_dir), ""] + sorted(os.listdir('.'))
                    self.is_dir = True
                else:
                    if is_micropython:
                        with open(fname) as f:
                            self.content = f.readlines()
                    else:
                        with open(fname, errors="ignore") as f:
                            self.content = f.readlines()
                    self.write_tabs = False
                    i = 0
                    for l in self.content:
                        self.content[i] = self.expandtabs(l.rstrip())
                        i += 1
            except OSError:
                self.message = "Error: file '" + fname + "' may not exist"
        self.hash = self.hash_buffer()
    def put_file(self, fname):
        tmpfile = fname + ".pyetmp"
        with open(tmpfile, "w") as f:
            for l in self.content:
                if self.write_tabs == 'y':
                    f.write(self.packtabs(l))
                else:
                    f.write(l)
                f.write('\n')
        try:
            os.remove(fname)
        except:
            pass
        os.rename(tmpfile, fname)
    def expandtabs(self, s):
        if '\t' in s:
            self.write_tabs = True
            sb = StringIO()
            pos = 0
            for c in s:
                if c == '\t':
                    sb.write(" " * (8 - pos % 8))
                    pos += 8 - pos % 8
                else:
                    sb.write(c)
                    pos += 1
            return sb.getvalue()
        else:
            return s
def pye_edit(content, tab_size=4, undo=50, io_device=None):
    if io_device is None:
        print("IO device not defined")
        return
    gc.collect()
    index = 0
    undo = max(4, (undo if type(undo) is int else 0))
    current_dir = os.getcwd()
    if content:
        slot = []
        for f in content:
            slot.append(Editor(tab_size, undo, io_device))
            if type(f) == str and f:
                try:
                    slot[index].get_file(f)
                except Exception as err:
                    slot[index].message = "{!r}".format(err)
            else:
                try:
                    slot[index].content = [str(_) for _ in f]
                except:
                    slot[index].content = [str(f)]
            index += 1
    else:
        slot = [Editor(tab_size, undo, io_device)]
        slot[0].get_file(current_dir)
    while True:
        try:
            index %= len(slot)
            key = slot[index].edit_loop()
            if key == KEY_QUIT:
                if len(slot) == 1:
                    break
                del slot[index]
            elif key == KEY_GET:
                f = slot[index].line_edit("Open file: ", "", Editor.file_char)
                if f is not None:
                    slot.append(Editor(tab_size, undo,io_device))
                    index = len(slot) - 1
                    slot[index].get_file(f)
            elif key == KEY_NEXT:
                index += 1
            elif key == KEY_FORCE_QUIT:
                break
        except Exception as err:
            slot[index].message = "{!r}".format(err)
    Editor.yank_buffer = []
    os.chdir(current_dir)
    return slot[0].content if (slot[0].fname == "") else slot[0].fname
try:
    import usys as sys
except:
    import sys
class IO_DEVICE:
    def __init__(self):
        try:
            from micropython import kbd_intr
            kbd_intr(-1)
        except ImportError:
            pass
        if hasattr(sys.stdin, "buffer"):
            self.rd_raw_fct = sys.stdin.buffer.read
        else:
            self.rd_raw_fct = sys.stdin.read
    def wr(self, s):
        sys.stdout.write(s)
    def rd(self):
        return sys.stdin.read(1)
    def rd_raw(self):
        return self.rd_raw_fct(1)
    def deinit_tty(self):
        try:
            from micropython import kbd_intr
            kbd_intr(3)
        except ImportError:
            pass
    def get_screen_size(self):
        self.wr('\x1b[999;999H\x1b[6n')
        pos = ''
        char = self.rd()
        while char != 'R':
            pos += char
            char = self.rd()
        return [int(i, 10) for i in pos.lstrip("\n\x1b[").split(';')]
if "pye_edit" not in globals().keys():
    from pye import pye_edit
def pye(*args, tab_size=4, undo=50):
    io_device = IO_DEVICE()
    ret = pye_edit(args, tab_size=tab_size, undo=undo, io_device=io_device)
    io_device.deinit_tty()
    return ret



"""Implements a HD44780 character LCD connected via PCF8574 on I2C.
   This was tested with: https://www.wemos.cc/product/d1-mini.html"""

from lcd_api import LcdApi
import busio
from time import sleep

# The PCF8574 has a jumper selectable address: 0x20 - 0x27
DEFAULT_I2C_ADDR = 0x27

# Defines shifts or masks for the various LCD line attached to the PCF8574

MASK_RS = 0x01
MASK_RW = 0x02
MASK_E = 0x04
SHIFT_BACKLIGHT = 3
SHIFT_DATA = 4


class I2cLcd(LcdApi):
    """Implements a HD44780 character LCD connected via PCF8574 on I2C."""
    def __init__(self, i2c, i2c_addr, num_lines, num_columns):
        self.i2c = i2c
        self.i2c_addr = i2c_addr
        self.i2c.writeto(self.i2c_addr, bytearray([0]))
        sleep(0.02)   # Allow LCD time to powerup
		# Send reset 3 times
        self.hal_write_init_nibble(self.LCD_FUNCTION_RESET)
        sleep(0.005)    # need to delay at least 4.1 msec
        self.hal_write_init_nibble(self.LCD_FUNCTION_RESET)
        sleep(0.001)
        self.hal_write_init_nibble(self.LCD_FUNCTION_RESET)
        sleep(0.001)
        # Put LCD into 4 bit mode
        self.hal_write_init_nibble(self.LCD_FUNCTION)
        sleep(0.001)
        LcdApi.__init__(self, num_lines, num_columns)
        cmd = self.LCD_FUNCTION
        if num_lines > 1:
            cmd |= self.LCD_FUNCTION_2LINES
        self.hal_write_command(cmd)

    def hal_write_init_nibble(self, nibble):
        """Writes an initialization nibble to the LCD.

        This particular function is only used during initialization.
        """
        byte = ((nibble >> 4) & 0x0f) << SHIFT_DATA
        self.i2c.writeto(self.i2c_addr, bytearray([byte | MASK_E]))
        self.i2c.writeto(self.i2c_addr, bytearray([byte]))

    def hal_backlight_on(self):
        """Allows the hal layer to turn the backlight on."""
        self.i2c.writeto(self.i2c_addr, bytearray([1 << SHIFT_BACKLIGHT]))

    def hal_backlight_off(self):
        """Allows the hal layer to turn the backlight off."""
        self.i2c.writeto(self.i2c_addr, bytearray([0]))

    def hal_write_command(self, cmd):
        """Writes a command to the LCD.

        Data is latched on the falling edge of E.
        """
        byte = ((self.backlight << SHIFT_BACKLIGHT) | (((cmd >> 4) & 0x0f) << SHIFT_DATA))
        self.i2c.writeto(self.i2c_addr, bytearray([byte | MASK_E]))
        self.i2c.writeto(self.i2c_addr, bytearray([byte]))
        byte = ((self.backlight << SHIFT_BACKLIGHT) | ((cmd & 0x0f) << SHIFT_DATA))
        self.i2c.writeto(self.i2c_addr, bytearray([byte | MASK_E]))
        self.i2c.writeto(self.i2c_addr, bytearray([byte]))
        if cmd <= 3:
            # The home and clear commands require a worst case delay of 4.1 msec
            sleep(0.005)

    def hal_write_data(self, data):
        """Write data to the LCD."""
        byte = (MASK_RS | (self.backlight << SHIFT_BACKLIGHT) | (((data >> 4) & 0x0f) << SHIFT_DATA))
        self.i2c.writeto(self.i2c_addr, bytearray([byte | MASK_E]))
        self.i2c.writeto(self.i2c_addr, bytearray([byte]))
        byte = (MASK_RS | (self.backlight << SHIFT_BACKLIGHT) | ((data & 0x0f) << SHIFT_DATA))
        self.i2c.writeto(self.i2c_addr, bytearray([byte | MASK_E]))
        self.i2c.writeto(self.i2c_addr, bytearray([byte]))

    def hal_sleep_us(self, usecs):
        """Sleep for some time (given in microseconds)."""
        sleep(float(usecs)/1e6)


#! /usr/bin/python

# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""This class implements a lexical analyser capable
of consuming BASIC statements and commands and returning
a corresponding list of tokens.

>>> lexer = Lexer()
>>> tokenlist = lexer.tokenize('100 LET I = 10')
>>> tokenlist[0].pretty_print()
Column: 0 Category: UNSIGNEDINT Lexeme: 100
>>> tokenlist = lexer.tokenize('100 IF I <> 10')
>>> tokenlist[3].pretty_print()
Column: 9 Category: NOTEQUAL Lexeme: <>
>>> tokenlist = lexer.tokenize('100 LET I = 3.45')
>>> tokenlist[4].pretty_print()
Column: 12 Category: UNSIGNEDFLOAT Lexeme: 3.45
>>> tokenlist = lexer.tokenize('100 LET I = "HELLO"')
>>> tokenlist[4].pretty_print()
Column: 12 Category: STRING Lexeme: HELLO
"""

from basictoken import BASICToken as Token


class Lexer:

    def __init__(self):

        self.__column = 0  # Current column number
        self.__stmt = ''   # Statement string being processed

    def tokenize(self, stmt):
        """Returns a list of tokens obtained by
        lexical analysis of the specified
        statement.

        """
        self.__stmt = stmt
        self.__column = 0

        # Establish a list of tokens to be
        # derived from the statement
        tokenlist = []

        # Process every character until we
        # reach the end of the statement string
        c = self.__get_next_char()
        while c != '':

            # Skip any preceding whitespace
            while c.isspace():
                c = self.__get_next_char()

            # Construct a token, column count already
            # incremented
            token = Token(self.__column - 1, None, '')

            # Process strings
            if c == '"':
                token.category = Token.STRING

                # Consume all of the characters
                # until we reach the terminating
                # quote. Do not store the quotes
                # in the lexeme
                c = self.__get_next_char()  # Advance past opening quote

                # We explicitly support empty strings
                if c == '"':
                    # String is empty, leave lexeme as ''
                    # and advance past terminating quote
                    c = self.__get_next_char()

                else:
                    while True:
                        token.lexeme += c  # Append the current char to the lexeme
                        c = self.__get_next_char()

                        if c == '':
                            raise SyntaxError("Mismatched quotes")

                        if c == '"':
                            c = self.__get_next_char()  # Advance past terminating quote
                            break

            # Process numbers
            elif c.isdigit():
                token.category = Token.UNSIGNEDINT
                found_point = False

                # Consume all of the digits, including any decimal point
                while True:
                    token.lexeme += c  # Append the current char to the lexeme
                    c = self.__get_next_char()

                    # Break if next character is not a digit
                    # and this is not the first decimal point
                    if not c.isdigit():
                        if c == '.':
                            if not found_point:
                                found_point = True
                                token.category = Token.UNSIGNEDFLOAT

                            else:
                                # Another decimal point found
                                break

                        else:
                            break

            # Process keywords and names
            elif c.isalpha():
                # Consume all of the letters
                while True:
                    token.lexeme += c  # append the current char to the lexeme
                    c = self.__get_next_char()

                    # Break if not a letter or a dollar symbol
                    # (the latter is used for string variable names)
                    if not (c.isdigit() or c.isalpha() or c == '_' or c == '$'):
                        break

                # Normalise keywords and names to upper case
                token.lexeme = token.lexeme.upper()

                # Determine if the lexeme is a variable name or a
                # reserved word
                if token.lexeme in Token.keywords:
                    token.category = Token.keywords[token.lexeme]

                else:
                    token.category = Token.NAME

                # Remark Statments - process rest of statement without checks
                if token.lexeme == "REM":
                    while c!= '':
                        token.lexeme += c  # Append the current char to the lexeme
                        c = self.__get_next_char()

            # Process operator symbols
            elif c in Token.smalltokens:
                save = c
                c = self.__get_next_char()  # c might be '' (end of stmt)
                twochar = save + c

                if twochar in Token.smalltokens:
                    token.category = Token.smalltokens[twochar]
                    token.lexeme = twochar
                    c = self.__get_next_char() # Move past end of token

                else:
                    # One char token
                    token.category = Token.smalltokens[save]
                    token.lexeme = save

            # We do not recognise this token
            elif c != '':
                raise SyntaxError('Syntax error: Unrecognized Token')

            # Append the new token to the list
            tokenlist.append(token)

        return tokenlist

    def __get_next_char(self):
        """Returns the next character in the
        statement, unless the last character has already
        been processed, in which case, the empty string is
        returned.

        """
        if self.__column < len(self.__stmt):
            next_char = self.__stmt[self.__column]
            self.__column = self.__column + 1

            return next_char

        else:
            return ''


# SPDX-FileCopyrightText: 2020 Brent Rubell for Adafruit Industries
#
# SPDX-License-Identifier: MIT

import ipaddress
import ssl
import wifi
import socketpool
import adafruit_requests
from os import getenv

# URLs to fetch from
TEXT_URL = "http://wifitest.adafruit.com/testwifi/index.html"
JSON_QUOTES_URL = "https://www.adafruit.com/api/quotes.php"
JSON_STARS_URL = "https://api.github.com/repos/adafruit/circuitpython"

# Get wifi details and more from a .env file
if getenv('CIRCUITPY_WIFI_SSID') is None:
    raise Exception("WiFi secrets are kept in .env, please add them there!")

print("ESP32-S2 WebClient Test")

print("My MAC addr:", [hex(i) for i in wifi.radio.mac_address])

print("Available WiFi networks:")
for network in wifi.radio.start_scanning_networks():
    print("\t%s\t\tRSSI: %d\tChannel: %d" % (str(network.ssid, "utf-8"),
            network.rssi, network.channel))
wifi.radio.stop_scanning_networks()

print("Connecting to %s" % getenv('CIRCUITPY_WIFI_SSID'))
wifi.radio.connect(getenv('CIRCUITPY_WIFI_SSID'), getenv('CIRCUITPY_WIFI_PASSWORD'))
print("Connected to %s!" % getenv('CIRCUITPY_WIFI_SSID'))
print("My IP address is", wifi.radio.ipv4_address)

ipv4 = ipaddress.ip_address("8.8.4.4")
print("Ping google.com: %f ms" % (wifi.radio.ping(ipv4)*1000))

pool = socketpool.SocketPool(wifi.radio)
requests = adafruit_requests.Session(pool, ssl.create_default_context())

print("Fetching text from", TEXT_URL)
response = requests.get(TEXT_URL)
print("-" * 40)
print(response.text)
print("-" * 40)

print("Fetching json from", JSON_QUOTES_URL)
response = requests.get(JSON_QUOTES_URL)
print("-" * 40)
print(response.json())
print("-" * 40)

print()

print("Fetching and parsing json from", JSON_STARS_URL)
response = requests.get(JSON_STARS_URL)
print("-" * 40)
print("CircuitPython GitHub Stars", response.json()["stargazers_count"])
print("-" * 40)
response.close()
requests._free_sockets()

print("done")



import machine, time

def iweekDay():
    offset = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
    month = time.localtime()[1]
    day = time.localtime()[2]
    year = time.localtime()[0]
    afterFeb = 1
    if month > 2: afterFeb = 0
    aux = year - 1700 - afterFeb
    # dayOfWeek for 1700/1/1 = 5, Friday
    dayOfWeek  = 5
    # partial sum of days betweem current date and 1700/1/1
    dayOfWeek += (aux + afterFeb) * 365
    # leap year correction
    dayOfWeek += aux / 4 - aux / 100 + (aux + 100) / 400
    # sum monthly and day offsets
    dayOfWeek += offset[month - 1] + (day - 1)
    dayOfWeek %= 7
    return int(dayOfWeek)


rtcBase=0x4005c000
atomicBSet=0x2000

print("The current time is: %2.2i:%2.2i:%2.2i" % (time.localtime()[3], time.localtime()[4], time.localtime()[5]))
newDate = input("Enter the new time (hh:mm:ss): ")

if newDate != "":
    inDate = newDate.split(':')

    if len(inDate) != 3:
        print("Bad time format")
    elif int(inDate[0]) not in range(0,25):
        print("Invalid hour entered (0-24)")
    elif int(inDate[1]) not in range(0,61):
        print("invalid minute entered (0-60)")
    elif int(inDate[2]) not in range(0,61):
        print("invalid seconds entered (0-60)")
    else:
        machine.mem32[rtcBase+8] = (int(inDate[0]) << 16) | (int(inDate[1]) << 8) | int(inDate[2]) | iweekDay() << 24
        machine.mem32[rtcBase+atomicBSet+0xc] = 0x10


#! /usr/bin/python

# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Class representing a BASIC program.
This is a list of statements, ordered by
line number.

"""

from basicdata import BASICData
from basictoken import BASICToken as Token
from basicparser import BASICParser
from flowsignal import FlowSignal
from lexer import Lexer
import gc
from os import listdir,remove
from sys import implementation
try:
    from pydos_ui import input
except:
    pass
gc.collect()

class Program:

    def __init__(self):
        # Dictionary to represent program
        # statements, keyed by line number
        self.__program = {}

        # Program counter
        self.__next_stmt = 0

        # Initialise return stack for subroutine returns
        self.__return_stack = []

        # return dictionary for loop returns
        self.__return_loop = {}

        # Setup DATA object
        self.__data = BASICData()

    def list(self, strt_line, end_line, infile, tmpfile):
        """Lists the program"""
        line_numbers = self.line_numbers()

        for line_number in line_numbers:
            if (int(line_number) >= strt_line and int(line_number) <= end_line) or strt_line == -1:
                print(line_number, end=' ')

                statement = self.getprogram(line_number,infile,tmpfile)
                for token in statement:
                    # Add in quotes for strings
                    if token.category == Token.STRING:
                        print('"' + token.lexeme + '"', end=' ')

                    else:
                        print(token.lexeme, end=' ')

                print()

    def save(self, file, infile, tmpfile):
        """Save the program

        :param file: The name and path of the save file

        """
        retCode = False

        ans = "Y"
        if file in listdir():
            ans = input("Overwrite "+file+" (y/n): ").upper()

        if ans == "Y":

            if file+".pYb" in listdir():
                remove(file+".pYb")

            try:
                with open(file+".pYb", 'w') as outfile:

                    line_numbers = self.line_numbers()

                    if file.split(".")[-1].upper() == "PGM":
                        filelen = 0
                        for line_number in line_numbers:
                            statement = self.getprogram(line_number,infile,tmpfile)
                            if len(statement) > 1 and statement[0].lexeme == "DATA":
                                sign = -1
                            else:
                                sign = 1
                            fileLine = str(line_number)+","+str(sign*self.__program[line_number])
                            outfile.write(fileLine+"\n")
                            filelen += (len(fileLine)+(0 if implementation.name.upper() in ['MICROPYTHON','CIRCUITPYTHON'] else 1))
                        outfile.write("-999,-999\n")
                        filelen += (10 if implementation.name.upper() in ['MICROPYTHON','CIRCUITPYTHON'] else 11)

                    for line_number in line_numbers:
                        fileLine = str(line_number)

                        statement = self.getprogram(line_number,infile,tmpfile)
                        for token in statement:
                            # Add in quotes for strings
                            if token.category == Token.STRING:
                                fileLine += ' "' + token.lexeme + '"'

                            else:
                                fileLine += " " + token.lexeme
                        outfile.write(fileLine+"\n")

                retCode = True

            except OSError:
                print("Could not save to file")

        return retCode

    def load(self, file, tmpfile):
        """Load the program

        :param file: The name and path of the file to be loaded
        tmpfile:     File handle for temporary basic workfile"""

        infile = None

        try:
            infile = open(file, 'r')
            fIndex = 0
            fOffset = 0
            pgmLoad = False
            if file.split(".")[-1].upper() == "PGM":
                pgmLoad = True
                for fileLine in infile:
                    fOffset += (len(fileLine) + (0 if implementation.name.upper() in ['MICROPYTHON','CIRCUITPYTHON'] else 1))
                    if len(fileLine) >= 9 and fileLine[0:9] == "-999,-999":
                        break

            infile.seek(0)
            for fileLine in infile:
                if pgmLoad:
                    line_number = int(fileLine.split(",")[0])
                    fIndex = int(fileLine.split(",")[1])
                    if len(fileLine) >= 9 and fileLine[0:9] == "-999,-999":
                        break
                    self.__program[line_number] = abs(fIndex)+fOffset
                    if fIndex < 0:
                        self.__data.addData(line_number,abs(fIndex)+fOffset)
                else:
                    if ((fileLine.strip()).replace("\n","")).replace("\r","") != "":
                        line_number = int(fileLine.strip().split(" ")[0])
                        self.__program[line_number] = fIndex+fOffset
                        if fileLine.strip().upper()[fileLine.strip().find(' '):].strip()[:4] == "DATA":
                            self.__data.addData(line_number,fIndex)
                        #self.add_stmt(Lexer().tokenize((fileLine.replace("\n","")).replace("\r","")),fIndex+fOffset,tmpfile)
                    fIndex += (len(fileLine) + (0 if implementation.name.upper() in ['MICROPYTHON','CIRCUITPYTHON'] else 1))


        except OSError:
            print("Could not read file")

        return infile

    def add_stmt(self, tokenlist, fIndex, tmpfile):
        """
        Adds the supplied token list
        to the program. The first token should
        be the line number. If a token list with the
        same line number already exists, this is
        replaced.

        :param tokenlist: List of BTokens representing a
        numbered program statement
        fIndex: if >= 0: location in loaded program file of statment
                if < 0: Indicates statement was not read and should
               be added to temporary basic workfile
        tmpfile: file handle of temporary basic workfile

        """
        try:
            line_number = int(tokenlist[0].lexeme)
            if fIndex >= 0:
                self.__program[line_number] = fIndex
                if tokenlist[1].lexeme == "DATA":
                    self.__data.addData(line_number,fIndex)
            else:
                tmpfile.seek(0)
                filelen = 0
                for lines in tmpfile:
                    filelen += (len(lines)+(0 if implementation.name.upper() in ['MICROPYTHON','CIRCUITPYTHON'] else 1))

                self.__program[line_number] = -(filelen+1)
                if tokenlist[1].lexeme == "DATA":
                    self.__data.addData(line_number,-(filelen+1))
                #self.__program[line_number] = -(len(tmpfile.read())+1)

                fileLine = str(line_number)
                for token in tokenlist[1:]:
                    # Add in quotes for strings
                    if token.category == Token.STRING:
                        fileLine += ' "' + token.lexeme + '"'
                    else:
                        fileLine += " " + token.lexeme
                tmpfile.write(fileLine+"\n")

        except TypeError as err:
            raise TypeError("Invalid line number: " +
                            str(err))

    def line_numbers(self):
        """Returns a list of all the
        line numbers for the program,
        sorted

        :return: A sorted list of
        program line numbers
        """
        line_numbers = list(self.__program.keys())
        line_numbers.sort()

        return line_numbers

    def getprogram(self, ln, infile, tmpfile):
        if self.__program[ln] >= 0:
            infile.seek(self.__program[ln])
            statement = Lexer().tokenize((infile.readline().strip().replace("\n","")).replace("\r",""))[1:]
        else:
            tmpfile.seek(-(self.__program[ln]+1))
            statement = Lexer().tokenize((tmpfile.readline().replace("\n","")).replace("\r",""))[1:]

        return statement

    def __execute(self, line_number, infile,tmpfile):
        """Execute the statement with the
        specified line number

        :param line_number: The line number

        :return: The FlowSignal to indicate to the program
        how to branch if necessary, None otherwise

        """
        if line_number not in self.__program.keys():
            raise RuntimeError("Line number " + line_number +
                               " does not exist")

        statement = self.getprogram(line_number,infile,tmpfile)

        number_of_stmts = 1
        for e in statement:
            if e.category == Token.COLON:
                number_of_stmts += 1
            elif e.category == Token.IF:
                # any colons after an IF statement are seperators for the THEN or ELSE clause
                # and will be processed by the recursive call to PARSE within the PARSE method
                break

        for cstmt_number in range(0,number_of_stmts):
            try:
            #if True:
                tmp_flow = self.__parser.parse(statement, line_number, cstmt_number, infile, tmpfile, self.__data)

            except RuntimeError as err:
                raise RuntimeError(str(err))

            except KeyboardInterrupt:
                return FlowSignal(ftype=FlowSignal.STOP)

            if tmp_flow:
                break

        return tmp_flow


    def execute(self,infile,tmpfile):
        """Execute the program"""

        self.__parser = BASICParser()

        self.__data.restore(0) # reset data pointer

        line_numbers = self.line_numbers()

        if len(line_numbers) > 0:
            # Set up an index into the ordered list
            # of line numbers that can be used for
            # sequential statement execution. The index
            # will be incremented by one, unless modified by
            # a jump
            index = 0
            self.set_next_line_number(line_numbers[index])

            # Run through the program until the
            # has line number has been reached
            while True:
                flowsignal = self.__execute(self.get_next_line_number(),infile,tmpfile)
                self.__parser.last_flowsignal = flowsignal


                if flowsignal:
                    if flowsignal.ftype == FlowSignal.SIMPLE_JUMP:
                        # GOTO or conditional branch encountered
                        try:
                            index = line_numbers.index(flowsignal.ftarget)

                        except ValueError:
                            raise RuntimeError("Invalid line number supplied in GOTO or conditional branch: "
                                               + str(flowsignal.ftarget)+ " in line " + str(self.get_next_line_number()))

                        self.set_next_line_number(flowsignal.ftarget)

                    elif flowsignal.ftype == FlowSignal.GOSUB:
                        # Subroutine call encountered
                        # Add line number of next instruction to
                        # the return stack
                        if index + 1 < len(line_numbers):
                            self.__return_stack.append(line_numbers[index + 1])

                        else:
                            raise RuntimeError("GOSUB at end of program, nowhere to return")

                        # Set the index to be the subroutine start line
                        # number
                        try:
                            index = line_numbers.index(flowsignal.ftarget)

                        except ValueError:
                            raise RuntimeError("Invalid line number supplied in subroutine call: "
                                               + str(flowsignal.ftarget))

                        self.set_next_line_number(flowsignal.ftarget)

                    elif flowsignal.ftype == FlowSignal.RETURN:
                        # Subroutine return encountered
                        # Pop return address from the stack
                        try:
                            index = line_numbers.index(self.__return_stack.pop())

                        except ValueError:
                            raise RuntimeError("Invalid subroutine return in line " +
                                               str(self.get_next_line_number()))

                        except IndexError:
                            raise RuntimeError("RETURN encountered without corresponding " +
                                               "subroutine call in line " + str(self.get_next_line_number()))

                        self.set_next_line_number(line_numbers[index])

                    elif flowsignal.ftype == FlowSignal.STOP:
                        break

                    elif flowsignal.ftype == FlowSignal.LOOP_BEGIN:
                        # Loop start encountered
                        # Put loop line number on the stack so
                        # that it can be returned to when the loop
                        # repeats
                        self.__return_loop[flowsignal.floop_var] = self.get_next_line_number()

                        # Continue to the next statement in the loop
                        index = index + 1

                        if index < len(line_numbers):
                            self.set_next_line_number(line_numbers[index])

                        else:
                            # Reached end of program
                            raise RuntimeError("Program terminated within a loop")

                    elif flowsignal.ftype == FlowSignal.LOOP_SKIP:
                        # Loop variable has reached end value, so ignore
                        # all statements within loop and move past the corresponding
                        # NEXT statement
                        index = index + 1
                        while index < len(line_numbers):
                            next_line_number = line_numbers[index]
                            #temp_tokenlist = self.__program[next_line_number]
                            temp_tokenlist = self.getprogram(next_line_number,infile,tmpfile)

                            if temp_tokenlist[0].category == Token.NEXT and \
                               len(temp_tokenlist) > 1:
                                # Check the loop variable to ensure we have not found
                                # the NEXT statement for a nested loop
                                if temp_tokenlist[1].lexeme == flowsignal.ftarget:
                                    # Move the statement after this NEXT, if there
                                    # is one
                                    index = index + 1
                                    if index < len(line_numbers):
                                        next_line_number = line_numbers[index]  # Statement after the NEXT
                                        self.set_next_line_number(next_line_number)
                                        break

                            index = index + 1

                        # Check we have not reached end of program
                        if index >= len(line_numbers):
                            # Terminate the program
                            break

                    elif flowsignal.ftype == FlowSignal.LOOP_REPEAT:
                        # Loop repeat encountered
                        # Pop the loop start address from the stack
                        try:
                            index = line_numbers.index(self.__return_loop.pop(flowsignal.floop_var))

                        except ValueError:
                            raise RuntimeError("Invalid loop exit in line " +
                                               str(self.get_next_line_number()))

                        except KeyError:
                            raise RuntimeError("NEXT encountered without corresponding " +
                                               "FOR loop in line " + str(self.get_next_line_number()))

                        self.set_next_line_number(line_numbers[index])

                else:
                    index = index + 1

                    if index < len(line_numbers):
                        self.set_next_line_number(line_numbers[index])

                    else:
                        # Reached end of program
                        break

        else:
            raise RuntimeError("No statements to execute")

    def delete(self):
        """Deletes the program by emptying the dictionary"""
        self.__program.clear()
        self.__data.delete()

    def delete_statement(self, line_number):
        """Deletes a statement from the program with
        the specified line number, if it exists

        :param line_number: The line number to be deleted

        """

        self.__data.delData(line_number)

        try:
            del self.__program[line_number]

        except KeyError:
            raise KeyError("Line number does not exist")

    def get_next_line_number(self):
        """Returns the line number of the next statement
        to be executed

        :return: The line number

        """

        return self.__next_stmt

    def set_next_line_number(self, line_number):
        """Sets the line number of the next
        statement to be executed

        :param line_number: The new line number

        """
        self.__next_stmt = line_number



try:
    from pydos_ui import input
except:
    pass

import supervisor,os

def runvm(runargv):
    def chkPath(tstPath):
        validPath = True

        simpPath = ""
        if tstPath == []:
            validPath = True
            simpPath = ""
        else:

            savDir = os.getcwd()

            for path in tstPath:
                if path == "":
                    os.chdir("/")

                elif os.getcwd() == "/" and path == "..":
                    validPath = False
                    break

                elif path == ".":
                    continue

                elif path == ".." and len(os.getcwd().split('/')) == 2:
                    os.chdir('/')

                elif path == "..":
                    os.chdir("..")

                elif path in os.listdir() and (os.stat(path)[0] & (2**15) == 0):
                    os.chdir(path)

                else:
                    validPath = False
                    simpPath = ""
                    break

            if validPath:
                simpPath = os.getcwd()
            os.chdir(savDir)

        return((validPath,simpPath))

    def absolutePath(argPath,currDir):

        if argPath[0] == '/':
            fullPath = argPath
        elif currDir == '/':
            fullPath = '/'+argPath
        else:
            fullPath = currDir+'/'+argPath

        if len(fullPath) > 1 and fullPath[-1] == '/':
            fullPath = fullPath[:-1]

        return(fullPath)


    args = runargv.split(" ")
    savDir = os.getcwd()
    args[0] = absolutePath(args[0],savDir)

    aPath = args[0].split("/")
    newdir = aPath.pop(-1)
    (validPath, tmpDir) = chkPath(aPath)
    if tmpDir == "" or tmpDir[-1] != "/":
        tmpDir += "/"

    if validPath:
        t = open('/PyD-code.py','w')
        r = open(tmpDir+newdir,'r')

        t.write('import os\nos.chdir("'+tmpDir[:(-1 if tmpDir != "/" else None)]+'")\n')
        t.write("global passedIn\n")
        t.write("passedIn = '"+" ".join(args[1:])+"'\n")
        t.write("global envVars\n")
        t.write("envVars = {}\n")
        for _ in envVars:
            t.write("envVars['"+_+"']='"+str(envVars[_]).replace("'",chr(92)+"'")+"'\n")

        t.write("__name__ = 'PyDOS'\n")
        t.write(r.read())
        t.write("\nos.chdir('/')\n")

        t.close()
        r.close()

        supervisor.set_next_code_file('/PyD-code.py',reload_on_success=True,reload_on_error=True)

        supervisor.reload()
    else:
        print("Invalid target path: "+args[0])

if __name__ != "PyDOS":
    passedIn = ""

if passedIn == "":
    passedIn = input("Python script to run: ")

runargv = passedIn
passedIn = ""

argline = runargv.split(" ")
runargv = argline[0]
if "." not in runargv:
    runargv = runargv + ".py"

if len(argline) > 1:
    runargv = runargv + " " + " ".join(argline[1:])

runvm(runargv)



from sys import implementation
from pydos_hw import Pydos_hw
import os

if implementation.name.upper() == "MICROPYTHON":
    from sys import print_exception
    from machine import Pin,SoftSPI,SPI
    try:
        from machine import SDCard
    except:
        pass
    import sdcard
elif implementation.name.upper() == "CIRCUITPYTHON":
    import adafruit_sdcard
    import storage


def sdMount(drive):
    
    def chkPath(tstPath):
        validPath = True

        simpPath = ""
        if tstPath == []:
            validPath = True
            simpPath = ""
        else:

            savDir = os.getcwd()

            for path in tstPath:
                if path == "":
                    os.chdir("/")

                elif os.getcwd() == "/" and path == "..":
                    validPath = False
                    break

                elif path == ".":
                    continue

                elif path == ".." and len(os.getcwd().split('/')) == 2:
                    os.chdir('/')

                elif path == "..":
                    os.chdir("..")

                elif path in os.listdir() and (os.stat(path)[0] & (2**15) == 0):
                    os.chdir(path)

                else:
                    validPath = False
                    simpPath = ""
                    break

            if validPath:
                simpPath = os.getcwd()
            os.chdir(savDir)

        return((validPath,simpPath))

    def absolutePath(argPath,currDir):

        if argPath[0] == '/':
            fullPath = argPath
        elif currDir == '/':
            fullPath = '/'+argPath
        else:
            fullPath = currDir+'/'+argPath

        if len(fullPath) > 1 and fullPath[-1] == '/':
            fullPath = fullPath[:-1]

        return(fullPath)

    def do_mount(drive):
        sdMounted = False

        if implementation.name.upper() == "MICROPYTHON":
            if Pydos_hw.SD_SCK and not altSPI:
                try:
                    os.mount(SDCard(), drive)
                    sdMounted = True
                except:
                    pass

                if not sdMounted:
                    try:
                        sd = sdcard.SDCard(Pydos_hw.SD_SPI(), Pin(Pydos_hw.SD_CS,Pin.OUT))
                        os.mount( sd, drive)
                        sdMounted = True
                    except Exception as e:
                        print_exception(e)
            else:
                try:
                    sd = sdcard.SDCard(Pydos_hw.SPI(), Pin(Pydos_hw.CS,Pin.OUT))
                    os.mount( sd, drive)
                    sdMounted = True
                except Exception as e:
                    print_exception(e)

        elif implementation.name.upper() == "CIRCUITPYTHON":
            if not Pydos_hw.SD_CS and not Pydos_hw.CS:
                print("CS Pin not allocated for SDCard SPI interface")
            else:
                try:
                    if altSPI:
                        sd = adafruit_sdcard.SDCard(Pydos_hw.SPI(), Pydos_hw.CS)
                    else:
                        if Pydos_hw.SD_CS:
                            sd = adafruit_sdcard.SDCard(Pydos_hw.SD_SPI(), Pydos_hw.SD_CS)
                        elif Pydos_hw.CS:
                            sd = adafruit_sdcard.SDCard(Pydos_hw.SD_SPI(), Pydos_hw.CS)
                    vfs = storage.VfsFat(sd)
                    storage.mount(vfs, drive)
                    sdMounted = True
                except Exception as e:
                    print('SD-Card: Fail,', e)

        if sdMounted:
            print(drive+" mounted")
            # nano connect/Tennsy 4.1 are special cases becuase LED uses the SPI SCK pin
            if os.uname().machine in ["arduino_nano_rp2040_connect", \
                "Arduino Nano RP2040 Connect with RP2040", \
                "TinyPICO with ESP32-PICO-D4"] and not altSPI:

                envVars[".sd_drive"] = drive

            if os.uname().machine in ["Teensy 4.1 with IMXRT1062DVJ6A", \
                'Teensy 4.1 with MIMXRT1062DVJ6A'] and altSPI:

                envVars[".sd_drive"] = drive
                
        return

    savDir = os.getcwd()
    args = absolutePath(drive,savDir)

    aPath = drive.split("/")
    newdir = aPath.pop(-1)
    (validPath, tmpDir) = chkPath(aPath)
    if tmpDir == "" or tmpDir[-1] != "/":
        tmpDir += "/"

    if validPath:
        if newdir not in os.listdir(tmpDir[:(-1 if tmpDir != "/" else None)]):
            if (tmpDir+newdir)[1:].find('/') != -1:
                print("Target must be in root")
            else:
                do_mount(tmpDir+newdir)
        else:
            print("Target name already exists")
    else:
        print("Invalid path")

    return

drive = "/sd"
altSPI = False

if __name__ != "PyDOS":
    passedIn = ""
    envVars = {}

if passedIn != "":
    drive = passedIn.split(',')
    if len(drive) > 1:
        altSPI = True
    drive = drive[0]
    
sdMount(drive)



#! /usr/bin/python

# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""This class implements a BASIC interpreter that
presents a prompt to the user. The user may input
program statements, list them and run the program.
The program may also be saved to disk and loaded
again.

"""

from basictoken import BASICToken as Token
from lexer import Lexer
from program import Program
from sys import stderr,implementation
from os import listdir,rename,remove
import gc
try:
    from pydos_ui import input
except:
    pass

if implementation.name.upper() == 'MICROPYTHON':
    from sys import print_exception

gc.collect()
if 'threshold' in dir(gc):
    gc.threshold(gc.mem_free() // 4 + gc.mem_alloc())

def main():

    banner = (
"""
     PPPP   Y   Y  BBBB    AAA    SSSS   I    CCC
     P   P   Y Y   B   B  A   A  S       I   C
     P   P   Y Y   B   B  A   A  S       I   C
     PPPP     Y    BBBB   AAAAA  SSSS    I   C
u  u P        Y    B   B  A   A      S   I   C
u  u P        Y    B   B  A   A      S   I   C
uuuu P        Y    BBBB   A   A  SSSS    I    CCC
                    microPyBasic
""")

    print(banner)

    lexer = Lexer()
    program = Program()

    fOpened = False
    #initialize and open temporary file
    tmpfile = open('_pybTmp.tmp','w+')
    infile = tmpfile

    #Attempting memory pre-allocation to prepare for large Basic programs
    if implementation.name.upper() in ['MICROPYTHON','CIRCUITPYTHON']:
        gc.collect()
        for i in range(1600):
            if i % 100 == 0:
                print(".",end="")
            try:
                program.__program[i] = i
            except:
                print("\nMemory pre-allocation limit reached at ",i)
                break
        print()
        program.__program.clear()

    if passedIn != "":
        infile = program.load(passedIn,tmpfile)
        program.execute(infile,tmpfile)

    # Continuously accept user input and act on it until
    # the user enters 'EXIT'
    while True:

        # kfw stmt = input_keyboard(': ')
        stmt = input(': ')

        try:
        #if True:
            tokenlist = lexer.tokenize(stmt)

            # Execute commands directly, otherwise
            # add program statements to the stored
            # BASIC program

            if len(tokenlist) > 0:

                # remove blank tokens
                i = 0
                iend = len(tokenlist)
                for _ in range(iend):
                    if i>0 and tokenlist[i].lexeme.strip() == "" and tokenlist[i].category != Token.STRING:
                        tokenlist.pop(i)
                    else:
                        i+=1

                # Exit the interpreter
                if tokenlist[0].category == Token.EXIT:
                    if fOpened:
                        infile.close()
                    tmpfile.close()
                    remove('_pybTmp.tmp')
                    break

                # Add a new program statement, beginning
                # a line number
                elif tokenlist[0].category == Token.UNSIGNEDINT\
                     and len(tokenlist) > 1:
                    program.add_stmt(tokenlist,-1,tmpfile)

                # Delete a statement from the program
                elif tokenlist[0].category == Token.UNSIGNEDINT \
                        and len(tokenlist) == 1:
                    program.delete_statement(int(tokenlist[0].lexeme))

                # Execute the program
                elif tokenlist[0].category == Token.RUN:
                    try:
                        program.execute(infile,tmpfile)

                    except KeyboardInterrupt:
                        print("Program terminated")

                # List the program
                elif tokenlist[0].category == Token.LIST:
                    if len(tokenlist) == 2:
                        program.list(int(tokenlist[1].lexeme),int(tokenlist[1].lexeme),infile,tmpfile)
                    elif len(tokenlist) == 3:
                        program.list(int(tokenlist[1].lexeme),int(tokenlist[2].lexeme),infile,tmpfile)
                    elif len(tokenlist) == 4:
                        program.list(int(tokenlist[1].lexeme),int(tokenlist[3].lexeme),infile,tmpfile)
                    else:
                        program.list(-1,-1,infile,tmpfile)

                # Save the program to disk
                elif tokenlist[0].category == Token.SAVE:
                    if len(tokenlist) <= 1:
                        print("No filename specified")
                    elif "/" in tokenlist[1].lexeme or "\\" in tokenlist[1].lexeme or ":" in tokenlist[1].lexeme:
                        print("Can only save to current directory")
                    else:
                        if tokenlist[1].lexeme.split(".")[-1].upper() == "PGM":
                            filename = tokenlist[1].lexeme+".BAS"
                            if filename in listdir():
                                remove(filename)
                        else:
                            filename = tokenlist[1].lexeme

                        if program.save(filename,infile,tmpfile):
                            # Since we are running the program from the disk file "in place"
                            # the current program listing is contained only on disk in the
                            # loaded file (if one has been loaded) and the _pYbTmp.tmp working file
                            # the program.save function has to save the file to a temporary filename
                            # and we now have to replace the specified output filename with that
                            # temporary file. Finally we need to reload the saved file to initialize
                            # the in place files and index
                            #
                            # PGM files are worse since we use the internal file index keys we have to
                            # first save as a normal BAS file and reaload to eliminate all code from
                            # the temporary file
                            program.delete()
                            if fOpened:
                                infile.close()

                            if filename in listdir():
                                remove(filename)

                            rename(filename+".pYb",filename)

                            tmpfile.close()
                            tmpfile = open('_pybTmp.tmp','w+')
                            infile = program.load(filename,tmpfile)
                            if infile != None:

                                if tokenlist[1].lexeme.split(".")[-1].upper() == "PGM":

                                    filename = tokenlist[1].lexeme
                                    if program.save(filename,infile,tmpfile):

                                        if filename in listdir():
                                            remove(filename)

                                        rename(filename+".pYb",filename)

                                        #.PGM.BAS file now being used, need to close .PGM.BAS and reload .pgm file
                                        program.delete()
                                        infile.close()
                                        infile = program.load(filename,tmpfile)
                                        if infile != None:
                                            remove(filename+".BAS")
                                            fOpened = True
                                            print("Program written to file")
                                        else:
                                            # This should never happen, but just in case...
                                            fOpened = False
                                            print("Program saved but lost from active memory, ",end="")
                                            print("enter load command to re-load saved work")
                                else:
                                    fOpened = True
                                    print("Program written to file")
                            else:
                                # This should never happen, but just in case...
                                fOpened = False
                                if tokenlist[1].lexeme.split(".")[-1].upper() == "PGM":
                                    rename(filename,tokenlist[1].lexeme+".BAS")
                                    print("Texed program saved as "+tokenlist[1].lexeme+".BAS but lost from active memory, ",end="")
                                else:
                                    print("Program saved but lost from active memory, ",end="")
                                print("enter load command to re-load saved work")


                # Load the program from disk and/or delete the program from memory
                elif tokenlist[0].category == Token.LOAD or tokenlist[0].category == Token.NEW:
                    program.delete()
                    if fOpened:
                        infile.close()
                    fOpened = False

                    tmpfile.close()
                    tmpfile = open('_pybTmp.tmp','w+')

                    if tokenlist[0].category == Token.LOAD:

                        if len(tokenlist) > 1:
                            infile = program.load(tokenlist[1].lexeme,tmpfile)
                            if infile != None:
                                fOpened = True
                                print("Program read from file")
                            else:
                                fOpened = False
                        else:
                            print("Program file not found")

                # Unrecognised input
                else:
                    print("Unrecognized input")
                    for token in tokenlist:
                        token.print_lexeme()
                    print("")

        # Trap all exceptions so that interpreter
        # keeps running
        except Exception as e:
            if implementation.name.upper() == 'MICROPYTHON':
                #### print(e)
                print_exception(e)
            else:
                print(e)

if __name__ != "PyDOS":
    passedIn = ""

main()



# Scanner i2c en MicroPython | MicroPython i2c scanner
# Renvoi l'adresse en decimal et hexa de chaque device connecte sur le bus i2c
# Return decimal and hexa adress of each i2c device
# https://projetsdiy.fr - https://diyprojects.io (dec. 2017)

# If you run this and it seems to hang, try manually unlocking
# your I2C bus from the REPL with
#  >>> import board
#  >>> board.I2C().unlock()
from pydos_hw import Pydos_hw
from sys import implementation

i2c = Pydos_hw.I2C()

print('Scan i2c bus...')

if implementation.name.upper() == 'CIRCUITPYTHON':
    while not i2c.try_lock():
        pass

devices = i2c.scan()

if len(devices) == 0:
    print("No i2c device !")
else:
    print('i2c devices found:',len(devices))

for device in devices:
    print("Decimal address: ",device," | Hex address: ",hex(device))

if implementation.name.upper() == 'CIRCUITPYTHON':
    i2c.unlock()
    #i2c.deinit()



# PyDOS Board Configuration for 'Adafruit QT Py RP2040 with RP2040'

Pydos_pins = {
    'sndPin' : (29,"A0"),
    'neoPixel' : (12,None),
    'neoPixel_Pow' : (11,None),
    'I2C_NUM' : (1,None),
    'SCL' : (23,"SCL"),
    'SDA' : (22,"SDA"),
    'SCK' : (6,"SCK"),
    'MOSI' : (3,"MO"),
    'MISO' : (4,"MI"),
    'CS' : (26,"A3")
}



import sys
print("\nRestarting PyDOS....")
print("If PyDOS doesn't load, press Ctrl-D at the >>> REPL prompt\n\n")
if sys.implementation.name.upper() == "CIRCUITPYTHON":
    from supervisor import reload
    reload()
elif sys.implementation.name.upper() == "MICROPYTHON":
    sys.exit()


"""
    GPIO abstraction layer

"""
from sys import implementation
from pydos_bcfg import Pydos_pins
if implementation.name.upper() == "MICROPYTHON":
    from os import uname
    from machine import Pin
    from machine import SoftI2C as s_I2C
    from machine import I2C as m_I2C
    from machine import SoftSPI as s_SPI
    from machine import SPI as m_SPI

elif implementation.name.upper() == "CIRCUITPYTHON":
    try:
        from adafruit_bus_device.i2c_device import I2CDevice
    except:
        pass
    import digitalio
    import busio
    import board

    if board.board_id == "unexpectedmaker_feathers2":
        try:
            import kfw_s2_board as board
        except ImportError:
            pass
    elif board.board_id == "raspberry_pi_pico":
        try:
            import kfw_pico_board as board
            foundBoard = True
        except ImportError:
            foundBoard = False

        if not foundBoard:
            try:
                import cyt_mpp_board as board
            except ImportError:
                pass

class PyDOS_HW:

    _I2C = None
    _I2C_power = None
    _SPI = None
    _SD_SPI = None
    sndPin = None
    sndGPIO = None
    neoPixel = None
    neoPixel_Pow = None
    dotStar_Clock = None
    dotStar_Data = None
    led = None
    I2C_NUM = None
    SCL = None
    SDA = None
    SD_SPI_NUM = None
    SD_SCK = None
    SD_MOSI = None
    SD_MISO = None
    SD_CS = None
    SPI_NUM = None
    SCK = None
    MOSI = None
    MISO = None
    CS = None
    I2CbbqDevice = None
    KFW = False

    def __init__(self):

        self.sndPin = Pydos_pins.get('sndPin',(None,None))[0]
        self.neoPixel = Pydos_pins.get('neoPixel',(None,None))[0]
        self.neoPixel_Pow = Pydos_pins.get('neoPixel_Pow',(None,None))[0]
        self.dotStar_Clock = Pydos_pins.get('dotStar_Clock',(None,None))[0]
        self.dotStar_Data = Pydos_pins.get('dotStar_Data',(None,None))[0]
        self.led = Pydos_pins.get('led',(None,None))[0]

        if implementation.name.upper() == 'MICROPYTHON':
            if self.sndPin:
                self.sndPin = Pin(self.sndPin)

            if self.neoPixel:
                self.neoPixel = Pin(self.neoPixel)

            if self.neoPixel_Pow:
                self.neoPixel_Pow = Pin(self.neoPixel_Pow,Pin.OUT)

        self.I2C_NUM=Pydos_pins.get('I2C_NUM',(None,None))[0]
        self.SCL=Pydos_pins.get('SCL',(None,None))[0]
        self.SDA=Pydos_pins.get('SDA',(None,None))[0]
        self.SD_SPI_NUM=Pydos_pins.get('SD_SPI_NUM',(None,None))[0]
        self.SD_SCK=Pydos_pins.get('SD_SCK',(None,None))[0]
        self.SD_MOSI=Pydos_pins.get('SD_MOSI',(None,None))[0]
        self.SD_MISO=Pydos_pins.get('SD_MISO',(None,None))[0]
        self.SD_CS=Pydos_pins.get('SD_CS',(None,None))[0]
        self.SPI_NUM=Pydos_pins.get('SPI_NUM',(None,None))[0]
        self.SCK=Pydos_pins.get('SCK',(None,None))[0]
        self.MOSI=Pydos_pins.get('MOSI',(None,None))[0]
        self.MISO=Pydos_pins.get('MISO',(None,None))[0]
        self.CS=Pydos_pins.get('CS',(None,None))[0]

        if implementation.name.upper() == 'CIRCUITPYTHON':
            if self.CS:
                self.CS = digitalio.DigitalInOut(self.CS)
            if self.SD_CS:
                self.SD_CS = digitalio.DigitalInOut(self.SD_CS)

            if self.sndPin:
                self.sndGPIO = digitalio.DigitalInOut(self.sndPin)
                self.sndGPIO.direction = digitalio.Direction.OUTPUT
                self.sndGPIO.value = False

            try:
                chkForFile = open('/lib/kfw_pico_board.py','r')
                chkForFile.close()
                self.KFW = True
            except:
                pass
            if not self.KFW:
                try:
                    chkForFile = open('/lib/kfw_s2_board.py','r')
                    chkForFile.close()
                    self.KFW = True
                except:
                    pass

            if self.KFW:
                self.neoPixel = board.D11
            elif self.neoPixel is None and 'NEOPIXEL' in dir(board):
                self.neoPixel = board.NEOPIXEL

    def quietSnd(self):

        if implementation.name.upper() == "CIRCUITPYTHON":
            if self.sndPin:
                self.sndGPIO = digitalio.DigitalInOut(self.sndPin)
                #self.sndGPIO.direction = digitalio.Direction.OUTPUT
        return

    def I2C(self):

        if not self._I2C:
            if implementation.name.upper() == "CIRCUITPYTHON":

                if 'I2C_POWER_INVERTED' in dir(board) and not self._I2C_power:
                    self._I2C_power = digitalio.DigitalInOut(board.I2C_POWER_INVERTED)
                    self._I2C_power.direction = digitalio.Direction.OUTPUT
                    self._I2C_power.value = False

                if 'STEMMA_I2C' in dir(board):
                    self._I2C = board.STEMMA_I2C()
                elif 'I2C' in dir(board):
                    self._I2C = board.I2C()
                else:
                    self._I2C = busio.I2C(self.SCL, self.SDA)

                if self.KFW and not self.I2CbbqDevice:
                    self.I2CbbqDevice = I2CDevice(self._I2C, 0x1F)
            elif implementation.name.upper() == "MICROPYTHON":
                if self.I2C_NUM:
                    self._I2C = m_I2C(self.I2C_NUM,scl=Pin(self.SCL),sda=Pin(self.SDA))
                else:
                    self._I2C = s_I2C(scl=Pin(self.SCL),sda=Pin(self.SDA))

        return self._I2C

    def I2C_deinit(self):
        if implementation.name.upper() == "CIRCUITPYTHON":
            if self._I2C_power:
                self._I2C_power.deinit()
                self._I2C_power = None

            if self._I2C:
                self._I2C.deinit()
            self._I2C = None
            self.I2CbbqDevice = None

    def SD_deinit(self):
            if self._SD_SPI:
                if self._SPI == self._SD_SPI:
                    self._SPI = None
                self.SD_SPI().deinit()
                self._SD_SPI = None

    def SPI_deinit(self):
        if self._SPI:
            if self._SPI == self._SD_SPI:
                self._SD_SPI = None
            self.SPI().deinit()
            self._SPI = None

    def SPI(self):
        if not self._SPI:
            if implementation.name.upper() == "CIRCUITPYTHON":
                if 'SPI' in dir(board):
                    self._SPI = board.SPI()
                else:
                    if self.SCK:
                        self._SPI = busio.SPI(self.SCK, self.MOSI, self.MISO)
            elif implementation.name.upper() == "MICROPYTHON":
                if Pydos_hw.SPI_NUM != None:
                    self._SPI = m_SPI(Pydos_hw.SPI_NUM)
                else:
                    if self.SCK:
                        self._SPI = s_SPI(sck=Pin(Pydos_hw.SCK),mosi=Pin(Pydos_hw.MOSI), \
                            miso=Pin(Pydos_hw.MISO))
        return self._SPI

    def SD_SPI(self):
        if not self._SD_SPI:
            if implementation.name.upper() == "CIRCUITPYTHON":
                if 'SD_SPI' in dir(board):
                    self._SD_SPI = board.SD_SPI()
                else:
                    if self.SD_SCK:
                        trybitbangio = False
                        try:
                            self._SD_SPI = busio.SPI(self.SD_SCK, self.SD_MOSI, self.SD_MISO)
                        except ValueError:
                            trybitbangio = True
                        except:
                            print('SPI Create Fail,', e)

                        if trybitbangio:
                            try:
                                import bitbangio
                            except ImportError:
                                trybitbangio = False

                        if trybitbangio:
                            self._SD_SPI = bitbangio.SPI(self.SD_SCK, self.SD_MOSI, self.SD_MISO)
                    else:
                        self._SD_SPI = self.SPI()
            elif implementation.name.upper() == "MICROPYTHON":
                if Pydos_hw.SD_SPI_NUM != None:
                    self._SD_SPI = m_SPI(Pydos_hw.SD_SPI_NUM)
                else:
                    if self.SD_SCK:
                        self._SD_SPI = s_SPI(sck=Pin(Pydos_hw.SD_SCK),mosi=Pin(Pydos_hw.SD_MOSI), \
                            miso=Pin(Pydos_hw.SD_MISO))
                    else:
                        self._SD_SPI = self.SPI()

        return self._SD_SPI

Pydos_hw = PyDOS_HW()
sndPin = Pydos_hw.sndPin
sndGPIO = Pydos_hw.sndGPIO



import board
import busio


_SPI = None
_UART = None
_I2C = None


A0 = board.GP26
A1 = board.GP27
A2 = board.GP20
A3 = board.GP21
A4 = board.GP22
A5 = board.GP28
SCK = board.GP18
COPI = board.GP19
MOSI = board.GP19
CIPO = board.GP16
MISO = board.GP16
RX = board.GP1
TX = board.GP0
D14 = board.GP13
MISC = board.GP13
SCL = board.GP5
SDA = board.GP4
D5 = board.GP6
D6 = board.GP7
D9 = board.GP8
D10 = board.GP9
D11 = board.GP10
D12 = board.GP11
D13 = board.GP12

GP0 = board.GP0
GP1 = board.GP1
GP2 = board.GP2
GP3 = board.GP3
GP4 = board.GP4
GP5 = board.GP5
GP6 = board.GP6
GP7 = board.GP7
GP8 = board.GP8
GP9 = board.GP9
GP10 = board.GP10
GP11 = board.GP11
GP12 = board.GP12
GP13 = board.GP13
GP14 = board.GP14
GP15 = board.GP15
GP16 = board.GP16
GP17 = board.GP17
GP18 = board.GP18
GP19 = board.GP19
GP20 = board.GP20
GP21 = board.GP21
GP22 = board.GP22
GP23 = board.GP23
GP24 = board.GP24
GP25 = board.GP25
GP26 = board.GP26
GP27 = board.GP27
GP28 = board.GP28
GP26_A0 = board.GP26_A0
GP27_A1 = board.GP27_A1
GP28_A2 = board.GP28_A2
LED = board.LED
NEOPIXEL = D11
SMPS_MODE = board.SMPS_MODE
VBUS_SENSE = board.VBUS_SENSE
VOLTAGE_MONITOR = board.VOLTAGE_MONITOR
board_id = board.board_id

def SPI():
    global _SPI

    if not _SPI:
        _SPI = busio.SPI(SCK, COPI, CIPO)

    return _SPI


def UART():
    global _UART

    if not _UART:
        _UART = busio.UART(TX, RX)

    return _UART


def I2C():
    global _I2C

    if not _I2C:
        _I2C = busio.I2C(SCL, SDA)

    return _I2C



# PyDOS Board Configuration for Unidentified board

"""
    sndPin
    neoPixel
    neoPixel_Pow
    dotStar_Clock
    dotStar_Data

    I2C_NUM
    SCL
    SDA

    SD_SPI_NUM
    SD_SCK
    SD_MOSI
    SD_MISO
    SD_CS

    SPI_NUM
    SCK
    MOSI
    MISO
    CS
"""
from sys import implementation

if implementation.name.upper() == "CIRCUITPYTHON":
    import board

if implementation.name.upper() == "MICROPYTHON":
    Pydos_pins = {
        'sndPin' : (7,"IO7"),
        'neoPixel' : (16,"IO16"),
        'I2C_NUM' : (1,None),
        'SCL' : (3,"SCL IO3"),
        'SDA' : (2,"SDA IO2"),
        'SD_SPI_NUM' : (1,None),
        'SCK' : (14,"SCK IO14"),
        'MOSI' : (15,"MO IO15"),
        'MISO' : (12,"MI IO12"),
        'CS' : (9,"SS IO9")
    }

elif implementation.name.upper() == "CIRCUITPYTHON":
    Pydos_pins = {}

    if 'BUZZER' in dir(board):
        Pydos_pins['sndPin'] = (board.BUZZER,"board.BUZZER")
    elif 'D12' in dir(board):
        Pydos_pins['sndPin'] = (board.D12,"D12 GPIO12")
    if 'NEOPIXEL' in dir(board):
        Pydos_pins['neoPixel'] = (board.NEOPIXEL,None)
    if 'NEOPIXEL_POWER' in dir(board):
        Pydos_pins['neoPixel_Pow'] = (board.NEOPIXEL_POWER,None)
    if 'DOTSTAR_CLOCK' in dir(board):
        Pydos_pins['dotStar_Clock'] = (board.DOTSTAR_CLOCK,None)
    if 'DOTSTAR_DATA' in dir(board):
        Pydos_pins['dotStar_Data'] = (board.DOTSTAR_DATA,None)
    if 'SCL' in dir(board):
        Pydos_pins['SCL'] = (board.SCL,"SCL")
        Pydos_pins['SDA'] = (board.SDA,"SDA")
    elif 'D3' in dir(board) and 'D2' in dir(board):
        Pydos_pins['SCL'] = (board.D3,"SCL D3")
        Pydos_pins['SDA'] = (board.D2,"SDA D2")


    if "SD_CS" in dir(board):
        Pydos_pins["SD_CS"] = (board.SD_CS,"SD_CS")
    if "CS" in dir(board):
        Pydos_pins["CS"] = (board.CS,"CS")
    else:
        if 'D5' in dir(board):
            Pydos_pins["CS"] = (board.D5,"D5")
        elif 'GP5' in dir(board):
            Pydos_pins["CS"] = (board.GP5,"GP5")
        elif 'IO5' in dir(board):
            Pydos_pins["CS"] = (board.IO5,"IO5")

    _SCK = False
    if 'SD_SCK' in dir(board):
        _SCK = board.SD_SCK
    elif 'SD_CLK' in dir(board):
        _SCK = board.SD_CLK
    if _SCK:
        if 'SD_MOSI' in dir(board):
            Pydos_pins["SD_SCK"] = (_SCK,"SD_SCK")
            Pydos_pins["SD_MOSI"] = (board.SD_MOSI,"SD_MOSI")
            Pydos_pins["SD_MISO"] = (board.SD_MISO,"SD_MISO")
        elif 'SD_COPI' in dir(board):
            Pydos_pins["SD_SCK"] = (_SCK,"SD_SCK")
            Pydos_pins["SD_MOSI"] = (board.SD_COPI,"SD_COPI")
            Pydos_pins["SD_MISO"] = (board.SD_CIPO,"SD_CIPO")

    _SCK = False
    if 'SCK' in dir(board):
        _SCK = board.SCK
    elif 'CLK' in dir(board):
        _SCK = board.CLK
    if _SCK:
        if 'MOSI' in dir(board):
            Pydos_pins["SCK"] = (_SCK,"SCK")
            Pydos_pins["MOSI"] = (board.MOSI,"MOSI")
            Pydos_pins["MISO"] = (board.MISO,"MISO")
        elif 'COPI' in dir(board):
            Pydos_pins["SCK"] = (_SCK,"SCK")
            Pydos_pins["MOSI"] = (board.SD_COPI,"COPI")
            Pydos_pins["MISO"] = (board.SD_CIPO,"CIPO")




#! /usr/bin/python

# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Class to represent a token for the BASIC
programming language. A token consists of
three items:

column      Column in which token starts
category    Category of the token
lexeme      Token in string form

"""


class BASICToken:

        """BASICToken categories"""

        EOF             = 0   # End of file
        LET             = 1   # LET keyword
        LIST            = 2   # LIST command
        PRINT           = 3   # PRINT command
        RUN             = 4   # RUN command
        FOR             = 5   # FOR keyword
        NEXT            = 6   # NEXT keyword
        IF              = 7   # IF keyword
        THEN            = 8   # THEN keyword
        ELSE            = 9   # ELSE keyword
        ASSIGNOP        = 10  # '='
        LEFTPAREN       = 11  # '('
        RIGHTPAREN      = 12  # ')'
        PLUS            = 13  # '+'
        MINUS           = 14  # '-'
        TIMES           = 15  # '*'
        DIVIDE          = 16  # '/'
        NEWLINE         = 17  # End of line
        UNSIGNEDINT     = 18  # Integer
        NAME            = 19  # Identifier that is not a keyword
        EXIT            = 20  # Used to quit the interpreter
        DIM             = 21  # DIM keyword
        GREATER         = 22  # '>'
        LESSER          = 23  # '<'
        STEP            = 24  # STEP keyword
        GOTO            = 25  # GOTO keyword
        GOSUB           = 26  # GOSUB keyword
        INPUT           = 27  # INPUT keyword
        REM             = 28  # REM keyword
        RETURN          = 29  # RETURN keyword
        SAVE            = 30  # SAVE command
        LOAD            = 31  # LOAD command
        NOTEQUAL        = 32  # '<>'
        LESSEQUAL       = 33  # '<='
        GREATEQUAL      = 34  # '>='
        UNSIGNEDFLOAT   = 35  # Floating point number
        STRING          = 36  # String values
        TO              = 37  # TO keyword
        NEW             = 38  # NEW command
        EQUAL           = 39  # '='
        COMMA           = 40  # ','
        STOP            = 41  # STOP keyword
        COLON           = 42  # ':'
        ON              = 43  # ON keyword
        POW             = 44  # Power function
        SQR             = 45  # Square root function
        ABS             = 46  # Absolute value function
        DIM             = 47  # DIM keyword
        RANDOMIZE       = 48  # RANDOMIZE keyword
        RND             = 49  # RND keyword
        ATN             = 50  # Arctangent function
        COS             = 51  # Cosine function
        EXP             = 52  # Exponential function
        LOG             = 53  # Natural logarithm function
        SIN             = 54  # Sine function
        TAN             = 55  # Tangent function
        DATA            = 56  # DATA keyword
        READ            = 57  # READ keyword
        INT             = 58  # INT function
        CHR             = 59  # CHR$ function
        ASC             = 60  # ASC function
        STR             = 61  # STR$ function
        MID             = 62  # MID$ function
        MODULO          = 63  # MODULO operator
        TERNARY         = 64  # TERNARY functions
        VAL             = 65  # VAL function
        LEN             = 66  # LEN function
        UPPER           = 67  # UPPER function
        LOWER           = 68  # LOWER function
        ROUND           = 69  # ROUND function
        MAX             = 70  # MAX function
        MIN             = 71  # MIN function
        INSTR           = 72  # INSTR function
        AND             = 73  # AND operator
        OR              = 74  # OR operator
        NOT             = 75  # NOT operator
        PI              = 76  # PI constant
        RNDINT          = 77  # RNDINT function
        SEMICOLON       = 78  # ";"
        TAB             = 79  # TAB function
        OPEN            = 80  # OPEN keyword
        HASH            = 81  # "#"
        CLOSE           = 82  # CLOSE keyword
        FSEEK           = 83  # FSEEK keyword
        RESTORE         = 84  # RESTORE keyword
        SOUND           = 85  # SOUND keyword

        # Displayable names for each token category
        catnames = ['EOF', 'LET', 'LIST', 'PRINT', 'RUN',
        'FOR', 'NEXT', 'IF', 'THEN', 'ELSE', 'ASSIGNOP',
        'LEFTPAREN', 'RIGHTPAREN', 'PLUS', 'MINUS', 'TIMES',
        'DIVIDE', 'NEWLINE', 'UNSIGNEDINT', 'NAME', 'EXIT',
        'DIM', 'GREATER', 'LESSER', 'STEP', 'GOTO', 'GOSUB',
        'INPUT', 'REM', 'RETURN', 'SAVE', 'LOAD',
        'NOTEQUAL', 'LESSEQUAL', 'GREATEQUAL',
        'UNSIGNEDFLOAT', 'STRING', 'TO', 'NEW', 'EQUAL',
        'COMMA', 'STOP', 'COLON', 'ON', 'POW', 'SQR', 'ABS',
        'DIM', 'RANDOMIZE', 'RND', 'ATN', 'COS', 'EXP',
        'LOG', 'SIN', 'TAN', 'DATA', 'READ', 'INT',
        'CHR', 'ASC', 'STR', 'MID', 'MODULO', 'TERNARY',
        'VAL', 'LEN', 'UPPER', 'LOWER', 'ROUND',
        'MAX', 'MIN', 'INSTR', 'AND', 'OR', 'NOT', 'PI',
        'RNDINT', 'SEMICOLON', 'TAB', 'OPEN', 'HASH',
        'CLOSE', 'FSEEK', 'RESTORE', 'SOUND']

        smalltokens = {'=': ASSIGNOP, '(': LEFTPAREN, ')': RIGHTPAREN,
                       '+': PLUS, '-': MINUS, '*': TIMES, '/': DIVIDE,
                       '\n': NEWLINE, '<': LESSER,
                       '>': GREATER, '<>': NOTEQUAL,
                       '<=': LESSEQUAL, '>=': GREATEQUAL, ',': COMMA,
                       ':': COLON, '%': MODULO, '!=': NOTEQUAL, ';': SEMICOLON,
                       '#': HASH}

        # Dictionary of BASIC reserved words
        keywords = {'LET': LET, 'LIST': LIST, 'PRINT': PRINT,
                    'FOR': FOR, 'RUN': RUN, 'NEXT': NEXT,
                    'IF': IF, 'THEN': THEN, 'ELSE': ELSE,
                    'EXIT': EXIT, 'DIM': DIM, 'STEP': STEP,
                    'GOTO': GOTO, 'GOSUB': GOSUB,
                    'INPUT': INPUT, 'REM': REM, 'RETURN': RETURN,
                    'SAVE': SAVE, 'LOAD': LOAD, 'NEW': NEW,
                    'STOP': STOP, 'TO': TO, 'ON':ON, 'POW': POW,
                    'SQR': SQR, 'ABS': ABS,
                    'RANDOMIZE': RANDOMIZE, 'RND': RND,
                    'ATN': ATN, 'COS': COS, 'EXP': EXP,
                    'LOG': LOG, 'SIN': SIN, 'TAN': TAN,
                    'DATA': DATA, 'READ': READ, 'INT': INT,
                    'CHR$': CHR, 'ASC': ASC, 'STR$': STR,
                    'MID$': MID, 'MOD': MODULO,
                    'IF$': TERNARY, 'IFF': TERNARY,
                    'VAL': VAL, 'LEN': LEN,
                    'UPPER$': UPPER, 'LOWER$': LOWER,
                    'ROUND': ROUND, 'MAX': MAX, 'MIN': MIN,
                    'INSTR': INSTR, 'END': STOP,
                    'AND': AND, 'OR': OR, 'NOT': NOT,
                    'PI': PI, 'RNDINT': RNDINT, 'TAB': TAB,
                    'OPEN': OPEN, 'CLOSE': CLOSE, 'FSEEK': FSEEK,
                    'RESTORE': RESTORE, 'SOUND': SOUND}

        # Functions
        functions = {ABS, ATN, COS, EXP, INT, LOG, POW, RND, SIN, SQR, TAN,
                     CHR, ASC, MID, TERNARY, STR, VAL, LEN, UPPER, LOWER,
                     ROUND, MAX, MIN, INSTR, PI, RNDINT, TAB}

        def __init__(self, column, category, lexeme):

            self.column = column      # Column in which token starts
            self.category = category  # Category of the token
            self.lexeme = lexeme      # Token in string form

        def pretty_print(self):
            """Pretty prints the token

            """
            print('Column:', self.column,
                  'Category:', self.catnames[self.category],
                  'Lexeme:', self.lexeme)

        def print_lexeme(self):
            print(self.lexeme, end=' ')


import sys
from time import sleep
from pydos_hw import Pydos_hw
from pydos_ui import Pydos_ui
try:
    from pydos_ui import input
except ImportError:
    pass

if sys.implementation.name.upper() == "CIRCUITPYTHON":
    import supervisor
    from circuitpython_i2c_lcd import I2cLcd
else:
    import lcd2004
    import uselect



def lcdScroll(argv):

    if argv == "":
        if envVars.get("stopthread","") == "":
            argv = input("Say what?: ")
        else:
            print('\nUse "passedIn" environment variable to define text for threaded run.')

    # The PCF8574 has a jumper selectable address: 0x20 - 0x27
    DEFAULT_I2C_ADDR = 0x27

    i2c = Pydos_hw.I2C()

    if sys.implementation.name.upper() == "CIRCUITPYTHON":

        # circuitpython seems to require locking the i2c bus
        while i2c.try_lock():
            pass

        if DEFAULT_I2C_ADDR not in i2c.scan():
            print("LCD not found at address: ",hex(DEFAULT_I2C_ADDR))
            i2c.unlock()
            return

        # 2 lines, 16 characters per line
        lcd = I2cLcd(i2c, DEFAULT_I2C_ADDR, 2, 16)

        # smiley faces as custom characters
        happy = bytearray([0x00,0x0A,0x00,0x04,0x00,0x11,0x0E,0x00])
        heart = bytearray([0x00,0x00,0x0A,0X15,0X11,0X0A,0X04,0X00])
        grin = bytearray([0x00,0x00,0x0A,0x00,0x1F,0x11,0x0E,0x00])
        lcd.custom_char(0, happy)
        lcd.custom_char(1, heart)
        lcd.custom_char(2, grin)

    else:

        ID=1
        lcd=lcd2004.lcd(DEFAULT_I2C_ADDR,i2c)
        lcd.lcd_backlight(True)
        lcd.lcd_clear()

    if argv == "":
        if sys.implementation.name.upper() == "CIRCUITPYTHON":
            lcd.backlight_off()
        else:
            lcd.lcd_backlight(False)
    else:

        print("\nPress q to quit or set stopthread=stop...")

        mess = " "*max(0,16-len(argv))+argv+" "*max(0,(14-max(0,16-len(argv))))
        if sys.implementation.name.upper() == "CIRCUITPYTHON":
            mess = mess.replace("<3",chr(1))
            mess = mess.replace(":)",chr(0))
            mess = mess.replace("(:",chr(0))
            mess = mess.replace(":-)",chr(0))
            mess = mess.replace("(-:",chr(0))
            mess = mess.replace(":D",chr(2))
        if len(argv) != 1:
            mess =  " " + mess
        for i in range(16):
            if ((" "*(15-i))+mess[0:i+1]).strip() != "":
                if sys.implementation.name.upper() == "CIRCUITPYTHON":
                    lcd.move_to(0,0)
                    lcd.putstr((" "*(15-i))+mess[0:i+1])
                else:
                    lcd.lcd_print((" "*(15-i))+mess[0:i+1],1,0)
                sleep(.25)

        i = 0
        lastCell = 15
        envVars["stopthread"] = "go"
        cmnd = ""

        while cmnd.upper() != "Q" and envVars.get("stopthread") == "go":

            if Pydos_hw.KFW:
                i2c.unlock()
            while Pydos_ui.serial_bytes_available():
                cmnd = Pydos_ui.read_keyboard(1)
                print(cmnd, end="", sep="")
                if cmnd in "qQ":
                    break
            if Pydos_hw.KFW:
                while i2c.try_lock():
                    pass

            i += 1
            if i >= len(mess):
                i = 0

            line = mess[i:min(i+16,len(mess))]+mess[0:max(0,(i+16)-len(mess))]
            #lcd.lcd_clear()
            #lcd.lcd_print(line,1,0)
            lastCell += 1
            if lastCell > 39:
                lastCell = 0
            # Scroll Left
            if sys.implementation.name.upper() == "CIRCUITPYTHON":
                lcd.hal_write_command(0x10 | 0x8)
                lcd.move_to(lastCell,0)
                lcd.putstr(line[-1])
            else:
                lcd.lcd_write(0x10 | 0x8)
                lcd.lcd_print(line[-1],1,lastCell)
            sleep(.35)

        if "stopthread" in envVars.keys():
            envVars.pop("stopthread")

    if sys.implementation.name.upper() == "CIRCUITPYTHON":
        i2c.unlock()
        #i2c.deinit()

if __name__ != "PyDOS":
    passedIn = ""
    global envVars
    envVars = {}

lcdScroll(passedIn)



# PyDOS Board Configuration for 'Raspberry Pi Pico with RP2040 with Cytron Maker Pi Pico'

Pydos_pins = {
    'sndPin' : (18,"GP18 Onboard Buzzer"),
    'neoPixel' : (28,None),
    'I2C_NUM' : (0,None),
    'SCL' : (3,"GP3 GROVE#2"),
    'SDA' : (2,"GP2 GROVE#2"),
    'SD_SPI_NUM' : (1,None),
    'SCK' : (10,"GP10 Onboard SD"),
    'MOSI' : (11,"GP11 Onboard SD"),
    'MISO' : (12,"GP12 Onboard SD"),
    'SD_CS' : (15,"GP15 Onboard SD")
}



# PyDOS Board Configuration for 'FeatherS2 with ESP32-S2'

Pydos_pins = {
    'sndPin' : (10,"IO10"),
    'led' : (13,"IO13"),
    'dotStar_Clock' : (45,None),
    'dotStar_Data' : (40,None),
    'I2C_NUM' : (1,None),
    'SCL' : (9,"SCL IO9"),
    'SDA' : (8,"SDA IO8"),
    'SD_SPI_NUM' : (2,None),
    'SCK' : (36,"SCK IO36"),
    'MOSI' : (35,"MO IO35"),
    'MISO' : (37,"MI IO37"),
    'CS' : (33,"IO33")
}



import sys
from pydos_ui import Pydos_ui
from pydos_hw import Pydos_hw

if sys.implementation.name.upper() == 'MICROPYTHON':
    import machine
    from time import ticks_ms,sleep
    from os import uname
elif sys.implementation.name.upper() == 'CIRCUITPYTHON':
    from time import sleep
    from supervisor import ticks_ms
    from pwmio import PWMOut
    from board import board_id

def piano():

    if sys.implementation.name.upper() == 'MICROPYTHON':
        pwm=machine.PWM(Pydos_hw.sndPin)
    elif sys.implementation.name.upper() == 'CIRCUITPYTHON':
        Pydos_hw.sndGPIO.deinit() # Workaround for ESP32-S2 GPIO issue
        pwm=PWMOut(Pydos_hw.sndPin, duty_cycle=0, frequency=440, variable_frequency=True)
        # Hack for Teensy 4.1 PWM bug
        pwm.deinit()
        pwm = PWMOut(Pydos_hw.sndPin,duty_cycle=0,frequency=440,variable_frequency=True)
        pwm.deinit()
        pwm = PWMOut(Pydos_hw.sndPin,duty_cycle=0,frequency=440,variable_frequency=True)

    print ("\nPress +/- to change volume, 'q' to quit...")

    volume = 400
    cmnd = None
    press = False
    noneAt = ticks_ms()
    firstNone = True
    note = 0
    while cmnd != "q":
        if Pydos_ui.serial_bytes_available() != 0:
            cmnd = Pydos_ui.read_keyboard(1)
            #print("->"+cmnd+"<- : ",ord(cmnd[0]))
            firstNone = True
            if cmnd=="h": # middle C
                note=int(261.6256+.5)
                press = True
            elif cmnd == "a": # E
                note=int(164.8138+.5)
                press = True
            elif cmnd == "s": # F
                note=int(174.6141+.5)
                press = True
            elif cmnd == "d": # G
                note=int(195.9977+.5)
                press = True
            elif cmnd == "r": # G sharp/A flat
                note=int(207.6523+.5)
                press = True
            elif cmnd == "f": # A
                note=int(220)
                press = True
            elif cmnd == "t": # A sharp/B flat
                note=int(233.0819+.5)
                press = True
            elif cmnd == "g": # B
                note=int(246.9417+.5)
                press = True
            elif cmnd == "u": # C sharp/D flat
                note=int(277.1826+.5)
                press = True
            elif cmnd == "j": # D
                note=int(293.6648+.5)
                press = True
            elif cmnd == "k": # E
                note=int(329.6276+.5)
                press = True
            elif cmnd == "l": # F
                note=int(349.2262+.5)
                press = True
            elif cmnd == "o": # F sharp/G flat
                note=int(369.9944+.5)
                press = True
            elif cmnd == ";": # G
                note=int(391.9954+.5)
                press = True
            elif cmnd == "q":
                break

            if cmnd == "+":
                volume += 100
            elif cmnd == "-":
                volume -= 100
        else:
            if firstNone:
                noneAt = ticks_ms()
                firstNone = False
            else:
                if ticks_ms() > noneAt+500:
                    firstNone = True
                    press = False

        if press:
            if sys.implementation.name.upper() == 'MICROPYTHON':
                pwm.freq(note)
                if 'duty_u16' in dir(pwm):
                    pwm.duty_u16(volume)
                else:
                    pwm.duty(int((volume/65535)*1023))
                if "ESP32" in uname().machine or "S2" in uname().machine:
                    sleep(.1)
            elif sys.implementation.name.upper() == 'CIRCUITPYTHON':
                pwm.frequency = note
                pwm.duty_cycle = volume
                if "s2" in board_id or "s3" in board_id:
                    sleep(.1)

            #time.sleep(.1)
            #cmnd = kbdInterrupt()
            #while cmnd != None:
                #cmnd = kbdInterrupt()
            #pressedat = time.time()
        else:
            if sys.implementation.name.upper() == 'MICROPYTHON':
                if 'duty_u16' in dir(pwm):
                    pwm.duty_u16(0)
                else:
                    pwm.duty(0)
            elif sys.implementation.name.upper() == 'CIRCUITPYTHON':
                pwm.duty_cycle = 0
            #print("Release")

    if sys.implementation.name.upper() == 'CIRCUITPYTHON':
        pwm.deinit()
        Pydos_hw.quietSnd() # Workaround for ESP32-S2 GPIO issue

if Pydos_hw.sndPin:
    piano()
else:
    print("Sound Pin not found")



import network
import socket

def http_get(url):
    _, _, host, path = url.split('/', 3)
    addr = socket.getaddrinfo(host, 80)[0][-1]
    s = socket.socket()
    s.connect(addr)
    s.send(bytes('GET /%s HTTP/1.0\r\nHost: %s\r\n\r\n' % (path, host), 'utf8'))
    response = ""
    while True:
        data = s.recv(100)
        if data:
            response = response + str(data, 'utf8')
        else:
            break
    s.close()

    return response

TEXT_URL = "http://wifitest.adafruit.com/testwifi/index.html"

# Read config
config = {}
envfound = True
try:
    envfile = open('/.env')
except:
    envfound = False

if envfound:
    for line in envfile:
        try:
            config[line.split('=')[0].strip()] = line.split('=')[1].strip()
        except:
            pass
    envfile.close()

# Check .env has updated credentials
if not envfound or config.get('CIRCUITPY_WIFI_SSID','') == '':
    assert False, ("/.env has not been updated with your unique keys and data")

# Create WiFi connection and turn it on
wlan = network.WLAN(network.STA_IF)
if not wlan.active():
    wlan.active(True)

    # Connect to WiFi router
    print ("Connecting to WiFi: {}".format( config['CIRCUITPY_WIFI_SSID'] ) )
    wlan.connect( config['CIRCUITPY_WIFI_SSID'], config['CIRCUITPY_WIFI_PASSWORD'])

# Wait until wifi is connected
while not wlan.isconnected():
    pass

print("Fetching text from", TEXT_URL)
r = http_get(TEXT_URL)
print("-" * 40)
print(r)
print("-" * 40)



import board
import busio


_SPI = None
_UART = None
_I2C = None


A0 = board.A0
A1 = board.A1
A2 = board.A2
A3 = board.A3

SCK = board.GP10
COPI = board.GP11
MOSI = board.GP11
CIPO = board.GP12
MISO = board.GP12

NEOPIXEL = board.GP28
SNDPIN = board.GP18

D5 = board.GP15

GP0 = board.GP0
GP1 = board.GP1
GP2 = board.GP2
GP3 = board.GP3
GP4 = board.GP4
GP5 = board.GP5
GP6 = board.GP6
GP7 = board.GP7
GP8 = board.GP8
GP9 = board.GP9
GP10 = board.GP10
GP11 = board.GP11
GP12 = board.GP12
GP13 = board.GP13
GP14 = board.GP14
GP15 = board.GP15
GP16 = board.GP16
GP17 = board.GP17
GP18 = board.GP18
GP19 = board.GP19
GP20 = board.GP20
GP21 = board.GP21
GP22 = board.GP22
GP23 = board.GP23
GP24 = board.GP24
GP25 = board.GP25
GP26 = board.GP26
GP27 = board.GP27
GP28 = board.GP28
GP26_A0 = board.GP26_A0
GP27_A1 = board.GP27_A1
GP28_A2 = board.GP28_A2
LED = board.LED
SMPS_MODE = board.SMPS_MODE
VBUS_SENSE = board.VBUS_SENSE
VOLTAGE_MONITOR = board.VOLTAGE_MONITOR
board_id = board.board_id

def SPI():
    global _SPI

    if not _SPI:
        _SPI = busio.SPI(SCK, COPI, CIPO)

    return _SPI

def I2C():
    global _I2C

    if not _I2C:
        _I2C = busio.I2C(GP3, GP2)

    return _I2C




# PyDOS Board Configuration for 'lilygo_ttgo_t8_s2_st7789'

import board

Pydos_pins = {
    'sndPin' : (board.IO2,"IO2"),
    'SCL' : (board.IO4,"IO4"),
    'SDA' : (board.IO3,"IO3"),
    'SD_SCK' : (board.SD_CLK,"SD_CLK IO12"),
    'SD_MOSI' : (board.SD_MOSI,"SD_MOSI IO11"),
    'SD_MISO' : (board.SD_MISO,"SD_MISO IO13"),
    'SD_CS' : (board.SD_CS,"SD_CS IO15 Internal"),
    'SCK' : (board.IO8,"IO8"),
    'MOSI' : (board.IO7,"IO7"),
    'MISO' : (board.IO6,"IO6"),
    'CS' : (board.IO5,"IO5")
}



from sys import implementation
from pydos_bcfg import Pydos_pins

if implementation.name.upper() == "CIRCUITPYTHON":
    import board
    if board.board_id == 'raspberry_pi_pico':
        try:
            import kfw_pico_board as board
        except:
            pass

        try:
            import cyt_mpp_board as board
        except:
            pass
    elif board.board_id == 'unexpectedmaker_feathers2':
        try:
            import kfw_s2_board as board
        except:
            pass
elif implementation.name.upper() == "MICROPYTHON":
    from os import uname


def printPinAssignments():

    if implementation.name.upper() == "CIRCUITPYTHON":
        print("Board ID:              ",board.board_id)
    elif implementation.name.upper() == "MICROPYTHON":
        print("Board ID:              ",uname().machine)

    if Pydos_pins.get('sndPin',(None,None))[1]:
        print("PyDOS Sound Pin:       ",Pydos_pins['sndPin'][1])

    if Pydos_pins.get('SCL',(None,None))[1]:
        print("SCL Pin:               ",Pydos_pins['SCL'][1])

    if Pydos_pins.get('SDA',(None,None))[1]:
        print("SDA Pin:               ",Pydos_pins['SDA'][1])

    if implementation.name.upper() == "CIRCUITPYTHON":
        if board.board_id == "cytron_maker_pi_rp2040":
            print('*** The "+" pins on the SERVO header put out 5V')
            print('Be sure to use a grove 3v3 pin when connecting 3v3 devices')
            print('The GPIO pins on the SERVO header use 3.3V logic')
            print("I2C                    ","Grove #2")
        elif 'STEMMA_I2C' in dir(board):
            print("I2C                    ","board.STEMMA_I2C")
        elif 'I2C' in dir(board):
            print("I2C                    ","board.I2C")

        if 'SD_SPI' in dir(board):
            print("SD_SPI                 ","board.SD_SPI()")
        if 'SPI' in dir(board):
            print("SPI                    ","board.SPI()")

    if Pydos_pins.get('SD_CS',(None,None))[1]:
        print("SD_CS Pin:             ",Pydos_pins['SD_CS'][1])

    if Pydos_pins.get('SD_SCK',(None,None))[1]:
        print("SD_SCK Pin:            ",Pydos_pins['SD_SCK'][1])

    if Pydos_pins.get('SD_MOSI',(None,None))[1]:
        print("SD_MOSI Pin:           ",Pydos_pins['SD_MOSI'][1])

    if Pydos_pins.get('SD_MISO',(None,None))[1]:
        print("SD_MISO Pin:           ",Pydos_pins['SD_MISO'][1])

    if Pydos_pins.get('CS',(None,None))[1]:
        print("CS Pin:                ",Pydos_pins['CS'][1])

    if Pydos_pins.get('SCK',(None,None))[1]:
        print("SCK Pin:               ",Pydos_pins['SCK'][1])

    if Pydos_pins.get('MOSI',(None,None))[1]:
        print("MOSI Pin:              ",Pydos_pins['MOSI'][1])

    if Pydos_pins.get('MISO',(None,None))[1]:
        print("MISO Pin:              ",Pydos_pins['MISO'][1])

    if implementation.name.upper() == "CIRCUITPYTHON":
        if board.board_id == 'arduino_nano_rp2040_connect':
            # For Gemma M0, Trinket M0, Metro M0/M4 Express, ItsyBitsy M0/M4 Express
            print("Read/Write Ground Pin: ","board.D2")
        elif board.board_id in ['raspberry_pi_pico','cytron_maker_pi_rp2040']:
            print("Read/Write Ground Pin: ","board.GP6")
        elif board.board_id == 'adafruit_qtpy_esp32c3':
            print("Read/Write Ground Pin: ","board.A1")
        elif board.board_id == 'unexpectedmaker_feathers2':
            print("Read/Write Ground Pin: ","board.IO1")
        elif board.board_id == 'adafruit_itsybitsy_rp2040':
            print("Read/Write Ground Pin: ","board.D7")
        else:
            if "D5" in dir(board):
                print("Read/Write Ground Pin: ","board.D5")
            elif "GP5" in dir(board):
                print("Read/Write Ground Pin: ","board.GP5")
            elif "IO5" in dir(board):
                print("Read/Write Ground Pin: ","board.IO5")
            elif "D6" in dir(board):
                print("Read/Write Ground Pin: ","board.D6")
            elif "GP6" in dir(board):
                print("Read/Write Ground Pin: ","board.GP6")
            elif "IO6" in dir(board):
                print("Read/Write Ground Pin: ","board.IO6")
            elif "D2" in dir(board):
                print("Read/Write Ground Pin: ","board.D2")
            elif "GP2" in dir(board):
                print("Read/Write Ground Pin: ","board.GP2")
            elif "IO2" in dir(board):
                print("Read/Write Ground Pin: ","board.IO2")
            else:
                print("Read/Write Ground Pin: ","None")
                print("To set the PyDOS FS readonly and give the host write access")
                print("the following command at the PyDOS prompt: fs ro")

        print()
        print('*The Read/Write Ground Pin will no longer function')
        print('after the PyDOS "fs" command has been used')


    return

printPinAssignments()




import network,gc
try:
    import usocket as _socket
except:
    import _socket
try:
    import ussl as ssl
except:
    import ssl
import time

config = {}
envfound = True
try:
    envfile = open('/.env')
except:
    envfound = False

if envfound:
    for line in envfile:
        try:
            config[line.split('=')[0].strip()] = line.split('=')[1].strip()
        except:
            pass
    envfile.close()

# Check .env has updated credentials
if not envfound or config.get('CIRCUITPY_WIFI_SSID','') == '':
    assert False, ("/.env has not been updated with your unique keys and data")

# AP info
SSID=config['CIRCUITPY_WIFI_SSID'] # Network SSID
KEY=config['CIRCUITPY_WIFI_PASSWORD']  # Network key

PORT = 443
#HOST = "money.cnn.com"
#HOST2 = "www.money.cnn.com/data/markets"
HOST = "finance.yahoo.com"

wlan = network.WLAN(network.STA_IF)
if not wlan.active():
    # Init wlan module and connect to network
    print("Trying to connect. Note this may take a while...")

    wlan.active(True)
    wlan.connect(SSID, KEY)

# Wait until wifi is connected
while not wlan.isconnected():
    pass


# We should have a valid IP now via DHCP
print("Wi-Fi Connected ", wlan.ifconfig())

# Get addr info via DNS
addr = _socket.getaddrinfo(HOST, PORT)[0][4]
print("Connecting to: ",addr)

# Create a new socket and connect to addr
#client = _socket.socket(_socket.AF_INET, _socket.SOCK_STREAM)
client = _socket.socket()
client.connect(addr)

client = ssl.wrap_socket(client)
print(client)

# Set timeout
#client.settimeout(3.0)

# Send HTTP request and recv response
print("Fetching text from %s/quote/%%5EIXIC" % HOST)
#client.write("GET /data/markets HTTP/1.1\r\nHost: %s\r\n\r\n"%(HOST))
client.write("GET /quote/%%5EIXIC HTTP/1.1\r\nHost: %s\r\n\r\n"%(HOST))

print("-" * 40)
response = client.read(800).decode('utf-8')
print("Text Response: ", response[0:800])
print("-" * 40)
print()

nasdaq = -1
strtime = time.ticks_ms()
timer = 0
while nasdaq == -1 and timer < 60000:
    gc.collect()
    #response = client.read(4096).decode('utf-8')
    encresponse = client.read(4096)
    decerror = False
    try:
        response = encresponse.decode()
    except:
        decerror = True
        
    if decerror:
        response = ""
        for chresp in encresponse:
            try:
                response += chr(chresp)
            except:
                pass
        
    #nasdaq = response.find('data-ticker-name="Nasdaq"')
    nasdaq = response.find('data-symbol="^IXIC" data-field="regularMarketChangePercent"')
    pct = response[nasdaq:].find('%)')
    pctst = response[nasdaq+pct-15:].find('>')
    pctend = response[nasdaq+pct:].find('<')
    
    timedelta = time.ticks_ms() - strtime
    if timedelta > 0:
        timer += timedelta
        strtime += timedelta
    else:
        strtime = time.ticks_ms()
    
if nasdaq == -1:
    print("Timeout fetching Web data")
else:
    print("Nasdaq: ",response[nasdaq+pct-14+pctst:nasdaq+pct+pctend])

# Close socket
#sslclient.close()
client.close()



import machine
import sys
import uselect
import time
import lcd2004
from pydos_hw import Pydos_hw

def displayTemp(passedIn):

    def kbdInterrupt():

        spoll = uselect.poll()
        spoll.register(sys.stdin,uselect.POLLIN)

        cmnd = sys.stdin.read(1) if spoll.poll(0) else None

        spoll.unregister(sys.stdin)

        return(cmnd)

    lcdavail = True

    try:
        lcd=lcd2004.lcd(39,Pydos_hw.I2C())
    except:
        lcdavail = False
        
    if lcdavail:
        lcd.lcd_backlight(True)
        lcd.lcd_clear()
        lcd.lcd_print("Temperature:",1,0)

    print("q to quit...")

    tempsensor = machine.ADC(4)
    conversionFact = 3.3 /65535

    if passedIn == "":
        avgCount = 50
    else:
        avgCount = int(passedIn)

    cmd = ""
    while cmd != "Q" and cmd != "q":

        tempread = 0
        for i in range(avgCount):
            tempread += tempsensor.read_u16()/avgCount

        currVoltage = tempread * conversionFact
        temp = 27 - ((currVoltage - 0.706)/0.001721)
        temp = (temp * 9 / 5) + 32

        print("Temperature:%6.2f deg F" % (temp))
        if lcdavail:
            lcd.lcd_print((" "*16),2,0)
            lcd.lcd_print(("%6.2f deg F" % (temp)),2,0)

        time.sleep(5)
        cmd = kbdInterrupt()

if __name__ != "PyDOS":
    passedIn = ""

displayTemp(passedIn)



# PyDOS Board Configuration for 'teensy41'

import board

Pydos_pins = {
    'sndPin' : (board.D36,"D36"),
    'SCL' : (board.SCL,"SCL D19 A5"),
    'SDA' : (board.SDA,"SDA D18 A4"),
    'SD_SCK' : (board.CLK,"CLK D44 Internal"),
    'SD_MOSI' : (board.CMD,"CMD D45 Internal"),
    'SD_MISO' : (board.DAT0,"DAT0 D43 Internal"),
    'SD_CS' : (board.DAT3,"DAT3 D46 Internal"),
    'SCK' : (board.SCK,"SCK D13"),
    'MOSI' : (board.MOSI,"MOSI D11"),
    'MISO' : (board.MISO,"MISO D12"),
    'CS' : (board.D10,"D10")
}



# PyDOS Board Configuration for 'unexpectedmaker_feathers3'

import board

Pydos_pins = {
    'sndPin' : (board.IO10,"IO10"),
    'neoPixel' : (board.NEOPIXEL,None),
    'SCL' : (board.SCL,"SCL IO9"),
    'SDA' : (board.SDA,"SDA IO8"),
    'SCK' : (board.SCK,"SCK IO36"),
    'MOSI' : (board.MOSI,"MO IO35"),
    'MISO' : (board.MISO,"MI IO37"),
    'CS' : (board.D5,"D5 IO33")
}



# PyDOS Board Configuration for 'ESP32S3 module (spiram) with ESP32S3' (DevkitC-1-N8R2)

Pydos_pins = {
    'sndPin' : (10,"GPIO10"),
    'neoPixel' : (48,"GPIO48"),
    'I2C_NUM' : (1,None),
    'SCL' : (7,"GPIO7"),
    'SDA' : (6,"GPIO6"),
    'SPI_NUM' : (2,None),
    'SCK' : (36,"GPIO36"),
    'MOSI' : (35,"GPIO35"),
    'MISO' : (37,"GPIO37"),
    'CS' : (9,"GPIO9")
}



import time
import sys
from pydos_hw import Pydos_hw

if sys.implementation.name.upper() == 'MICROPYTHON':
    from machine import Pin
    from os import uname

    # nano connect is special case becuase LED uses the SPI SCK pin
    if uname().machine in ['Arduino Nano RP2040 Connect with RP2040','Teensy 4.1 with MIMXRT1062DVJ6A']:

        if envVars.get('.sd_drive',None) != None:
            from os import umount
            drive = envVars['.sd_drive']
            try:
                umount(drive) # Nano Connect uses LED pin for SPI SCK
                print("Unmounting SD card because of shared SCK pin")
                envVars.pop('.sd_drive',None)
            except:
                pass
            
        if uname().machine == 'Teensy 4.1 with MIMXRT1062DVJ6A':
            Pydos_hw.SPI_deinit()
        else:
            Pydos_hw.SD_deinit()

    if Pydos_hw.led:
        led = Pin(Pydos_hw.led, Pin.OUT)
    else:
        led = Pin(25, Pin.OUT)

elif sys.implementation.name.upper() == 'CIRCUITPYTHON':
    import board
    from digitalio import DigitalInOut, Direction

    # nano connect/Tennsy 4.1 are special cases becuase LED uses the SPI SCK pin
    if board.board_id in ["arduino_nano_rp2040_connect","teensy41"]:

        if envVars.get('.sd_drive',None) != None:
            from storage import umount
            from pydos_hw import PyDOS_HW
            drive = envVars['.sd_drive']
            try:
                umount(drive) # Nano Connect uses LED pin for SPI SCK
                print("Unmounting SD card because of shared SCK pin")
                envVars.pop('.sd_drive',None)
            except:
                pass
        # If anything has used the SPI interface need to free up SCK
        if board.board_id == "teensy41":
            Pydos_hw.SPI_deinit()
        else:
            Pydos_hw.SD_deinit()

    # LED setup for onboard LED
    if Pydos_hw.led:
        led = DigitalInOut(Pydos_hw.led)
    elif 'LED1' in dir(board):
        led = DigitalInOut(board.LED1)
    else:
        led = DigitalInOut(board.LED)
    led.direction = Direction.OUTPUT

from pydos_ui import Pydos_ui

def blink():

    print("listening..., Enter q to quit")
    cmnd = ""
    blinkstate = True
    while cmnd.upper() != "Q":

        while Pydos_ui.serial_bytes_available():
            cmnd = Pydos_ui.read_keyboard(1)
            print(cmnd, end="", sep="")
            if cmnd in "qQ":
                break

        if sys.implementation.name.upper() == "MICROPYTHON":
            led.value(blinkstate)
        elif sys.implementation.name.upper() == "CIRCUITPYTHON":
            led.value = blinkstate

        blinkstate = not blinkstate
        time.sleep(1)

    if sys.implementation.name.upper() == "CIRCUITPYTHON":
        led.deinit()
    else:
        led.value(False)


blink()



# PyDOS Board Configuration for 'adafruit_feather_esp32s2'

import board

Pydos_pins = {
    'sndPin' : (board.D12,"GPIO12 D12"),
    'neoPixel' : (board.NEOPIXEL,None),
    'SCL' : (board.SCL,"GPIO4 SCL"),
    'SDA' : (board.SDA,"GPIO3 SDA"),
    'SCK' : (board.SCK,"GPIO36 SCK"),
    'MOSI' : (board.MOSI,"GPIO35 MO"),
    'MISO' : (board.MISO,"GPIO37 MI"),
    'SD_CS' : (board.D5,"GPIO5 D5")
}
